"""Registry for T2T models."""

import inspect
from dataclasses import dataclass
from typing import Callable, NamedTuple, Optional, Union

import catalogue

from databricks.kie.t2t_models.base_t2t_model import BaseT2TModel
from databricks.kie.t2t_schema import T2TSystemParams


@dataclass
class T2TModelConfig:
    """Model config for model in T2T model registry."""
    name: str
    is_default: bool
    model: BaseT2TModel


class DefaultModelConfig(NamedTuple):
    """Configuration for the default T2T model."""
    name: str
    factory: Callable[[T2TSystemParams], BaseT2TModel]


class T2TModelRegistry(catalogue.Registry):
    """A thin wrapper around catalogue.Registry to add T2T models."""

    def __init__(
        self,
        *args,
        **kwargs,
    ) -> None:
        super().__init__(*args, **kwargs)
        self.default_model: Optional[DefaultModelConfig] = None

    def __call__(self, name: str, is_default: bool = False) -> BaseT2TModel:
        """Register a model to the registry."""
        if is_default and self.default_model:
            raise ValueError("Only one default model can be registered")

        def _wrapper(cls):
            self.register(name, to_register=cls, is_default=is_default)
            return cls

        return _wrapper

    def register(self,
                 name: str,
                 *,
                 to_register: Union[Callable[[T2TSystemParams], BaseT2TModel], BaseT2TModel],
                 is_default: bool = False):
        """Register a model to the registry."""
        if is_default and self.default_model:
            raise ValueError("Only one default model can be registered")

        if isinstance(to_register, type):
            # if register is a class, then it must implement BaseT2TModel
            if not issubclass(to_register, BaseT2TModel):
                raise ValueError("If registering a class, it must implement BaseT2TModel")

            func = to_register.create_from_system_param
        else:
            func = to_register

        # Inspect function signature
        sig = inspect.signature(func)
        params = sig.parameters
        # Get non-bounded arguments (exclude self/cls)
        non_bounded_args = [
            p for p in params.values() if p.kind not in (p.VAR_POSITIONAL, p.VAR_KEYWORD, p.KEYWORD_ONLY)
        ]
        if len(non_bounded_args) != 1:
            raise ValueError(f"Function {func.__name__} must have exactly 1 non-bounded argument")
        # Check argument type annotation
        arg = non_bounded_args[0]
        if arg.annotation != T2TSystemParams:
            raise ValueError("Argument of registered function must be 1 argument of type T2TSystemParams")

        if is_default:
            self.default_model = DefaultModelConfig(name=name, factory=func)

        return super().register(name, func=func)


model_registry = T2TModelRegistry("t2t_model_registry")



"""Baseline text-to-text model that uses templated instruction only."""
from typing import Any, Optional

from databricks.kie.inference_utils import DEFAULT_NUM_RETRIES, get_llm_proxy_chat_completion_response
from databricks.kie.t2t_models.base_t2t_model import BaseT2TModel
from databricks.kie.t2t_models.t2t_model_registry import model_registry
from databricks.kie.t2t_schema import T2TSystemParams
from databricks.kie.t2t_utils import create_chat_completions_messages_from_instruction, create_chat_completions_request


@model_registry("t2t_baseline_model", is_default=False)
class T2TBaselineModel(BaseT2TModel):
    """A baseline text-to-text model that uses templated instruction only.

    Args:
        instruction (str): The instruction guiding the model's behavior.
        temperature (Optional[float]): The temperature for sampling from the model.
    """
    DEFAULT_MODEL_ID = "gpt-4o-mini-2024-07-18-text2text"

    def __init__(self, instruction: str, temperature: Optional[float] = None):
        self.instruction = instruction
        self.temperature = temperature

    def __call__(self, model_input: str) -> str:
        messages = create_chat_completions_messages_from_instruction(instruction=self.instruction, inp=model_input)
        req = create_chat_completions_request(messages)
        return get_llm_proxy_chat_completion_response(self.DEFAULT_MODEL_ID,
                                                      req,
                                                      self.temperature,
                                                      num_retries=DEFAULT_NUM_RETRIES)

    @staticmethod
    def create_from_system_param(system_param: T2TSystemParams, **kwargs: Any) -> Optional['T2TBaselineModel']:
        instruction = system_param.instruction
        return T2TBaselineModel(instruction=instruction, temperature=kwargs.get("temperature", None))



"""Inference utility functions to invoke LLM's using ChatCompletionsRequest."""
import json
from typing import TYPE_CHECKING, Any, Dict, List, Optional

import requests as request_lib
from pydantic import BaseModel
from tenacity import Retrying, retry_if_exception, stop_after_attempt, wait_exponential

from databricks.model_training.api.utils import get_host_and_token_from_env, normalize_table_name

if TYPE_CHECKING:
    from databricks.model_serving.types.pt_endpoint import BaseEndpoint, TileMTBatchEndpoint

try:
    from openai.lib._parsing._completions import type_to_response_format_param
    OPENAI_AVAILABLE = True
except ImportError:
    OPENAI_AVAILABLE = False

# TODO(jun.choi): Make client ID configurable per usecase.
CLIENT_ID = 'finetuning-t2t'

# Constants
MARKDOWN_JSON_PATTERN = r'(?s).*?```(?:json)?\s*(.*?)\s*```\s*.*'
DOCUMENT_TEMPLATE = "\\n\\nDocument:\\n\\n"

RETRY_WAIT_EXPONENTIAL_MULTIPLIER = 1
RETRY_WAIT_EXPONENTIAL_MIN = 1
RETRY_WAIT_EXPONENTIAL_MAX = 10
DEFAULT_NUM_RETRIES = 5


class LLMProxyRateLimitError(Exception):

    def __init__(self, message):
        super().__init__(message)


def clean_request_text(text_col: str) -> str:
    """Create SQL to clean request text.

    Args:
        text_col: Column containing text to clean
    """
    return rf"""REPLACE(REPLACE(REPLACE({text_col}, '"', ''), '“', ''), '”', '')"""


def format_json_for_sql(json_str: str) -> str:
    """
    Escapes a JSON string for safe inclusion in a SQL query string.
    - Replaces single quotes with two single quotes (SQL string escape)
    - Removes any existing backslashes used for escaping single quotes

    Args:
        json_str (str): The JSON string to be escaped.

    Returns:
        str: The escaped string ready for SQL insertion.
    """
    # Remove backslashes before single quotes (e.g., \' -> ')
    cleaned_str = json_str.replace(r"\'", "'")

    # Escape single quotes for SQL by doubling them
    escaped_str = cleaned_str.replace("'", "''")

    return escaped_str


def format_ai_query_tiles(
    source: str,
    endpoint: "TileMTBatchEndpoint",
    input_column: str = 'request',
    output_column: str = 'response',
    max_tokens: int = 1024,
    temperature: float = 0.3,
) -> str:

    # Handle table name quoting
    table_name = normalize_table_name(source)

    # Build query components
    clean_request = clean_request_text(input_column)

    return f"""
    SELECT 
        *,
        ai_query(
            '{endpoint.endpoint_name}',
            {clean_request},
            failOnError => true,
            modelParameters => named_struct('max_tokens', {max_tokens}, 'temperature', {temperature})
        ) AS {output_column}
    FROM {table_name}
    """


def format_ai_query(
    source: str,
    endpoint: "BaseEndpoint",
    system_prompt: str,
    input_column: str = 'request',
    output_column: str = 'response',
    max_tokens: int = 1024,
    temperature: float = 0.3,
    trailing_prompt: Optional[str] = None,
    response_format_str: Optional[str] = None,
) -> str:
    """Format SQL query for AI extraction.

    Args:
        source: Source table name
        endpoint: Endpoint for AI queries
        system_prompt: System prompt for extraction


    Example:
        >>> query = format_ai_query(
        ...     source="my_table",
        ...     endpoint=my_endpoint,
        ...     system_prompt="Extract entities",
        ...     max_tokens=2048,
        ...     temperature=0.5,
        ... )

    Returns:
        Formatted SQL query string
    """

    # Handle table name quoting
    table_name = normalize_table_name(source)

    # Handle trailing prompt
    trailing_prompt = trailing_prompt or system_prompt

    # Build query components
    model_params = f"named_struct('max_tokens', {max_tokens}, 'temperature', {temperature})"
    clean_request = clean_request_text(input_column)
    clean_response_format = format_json_for_sql(response_format_str) if response_format_str else None

    full_prompt = f"""
        system_prompt || 
        '{DOCUMENT_TEMPLATE}' ||
        {clean_request} ||
        '\\n\\n' || 
        trailing_prompt
    """

    # Build AI query string
    if clean_response_format is not None:
        return f"""
        SELECT 
            *,
            ai_query(
                '{endpoint.endpoint.name}',
                {full_prompt},
                failOnError => true,
                modelParameters => {model_params},
                responseFormat => '{clean_response_format}'
            ) AS {output_column}
        FROM {table_name}
        CROSS JOIN (
            SELECT 
                '{system_prompt}' AS system_prompt,
                '{trailing_prompt}' AS trailing_prompt
        )
        """
    return f"""
    SELECT 
        *,
        CASE 
            WHEN extracted_response != '' THEN extracted_response 
            ELSE raw_response 
        END AS {output_column}
    FROM (
        SELECT 
            *,
            ai_query(
                '{endpoint.endpoint.name}',
                {full_prompt},
                failOnError => true,
                modelParameters => {model_params}
            ) AS raw_response,
            regexp_extract(raw_response, r'{MARKDOWN_JSON_PATTERN}', 1) AS extracted_response
        FROM {table_name}
        CROSS JOIN (
            SELECT 
                '{system_prompt}' AS system_prompt,
                '{trailing_prompt}' AS trailing_prompt
        )
    )
    """


def _create_request_db(chat_completions, model_id, timeout=None, temperature=None):
    api_url, api_token = get_host_and_token_from_env()
    headers = {
        'Authorization': f'Bearer {api_token}',
        'Content-Type': 'application/json',
    }
    api_path = f'{api_url}/api/2.0/conversation/proxy/chat/completions'
    payload = {
        '@method': 'databricksChatCompletionRequest',
        'params': {
            **chat_completions
        },
        'model': model_id,
        'metadata': {
            'clientId': CLIENT_ID
        },
    }

    if temperature is not None:
        payload['params']['temperature'] = temperature

    return request_lib.post(api_path, headers=headers, json=payload, timeout=timeout)


def _create_request_ext(chat_completions, model_id, timeout=None, temperature=None):
    api_url, api_token = get_host_and_token_from_env()
    headers = {
        'Authorization': f'Bearer {api_token}',
        'Content-Type': 'application/json',
    }
    api_path = f'{api_url}/api/2.0/conversation/proxy/chat/completions'

    payload = {
        '@method': 'openAiServiceChatCompletionRequest',
        'params': {
            **chat_completions
        },
        'model': model_id,
        'apiVersion': '2024-09-01-preview',
        'metadata': {
            'clientId': CLIENT_ID
        },
    }

    if temperature is not None:
        payload['params']['temperature'] = temperature

    return request_lib.post(api_path, headers=headers, json=payload, timeout=timeout)


_MODEL_FN_DICT = {
    # Common / shared models
    'databricks-meta-llama-3-70b-instruct': _create_request_db,
    'llama-3-1-405b': _create_request_db,
    'gpt-4o-2024-08-06': _create_request_ext,
    'gpt-4o-mini-2024-07-18': _create_request_ext,
    'gpt-4-0613': _create_request_ext,
    'gpt-4-turbo-2024-04-09': _create_request_ext,
    'gpt-35-turbo-1106': _create_request_ext,

    # Client specific models
    'gpt-4o-2024-08-06-text2text': _create_request_ext,
    'gpt-4o-mini-2024-07-18-text2text': _create_request_ext,
}


def get_llm_proxy_chat_completion_response(model_id: str,
                                           req: Dict,
                                           timeout: Optional[float] = None,
                                           temperature: Optional[float] = None,
                                           num_retries: int = 0) -> str:
    """Retrieves the chat completion response from a specified language model.

        Args:
            model_id (str): The identifier of the language model to use.
            req (dict): The request payload to be sent to the language model.
            timeout (float, optional): The timeout for the request. Defaults to None.
            temperature (float, optional): The temperature for the request. Defaults to None.
            num_retries (int, optional): The number of retries for the request. Defaults to 0.

        Returns:
            str: The content of the chat completion response.
    """

    def _fn():
        if model_id not in _MODEL_FN_DICT:
            raise ValueError(f'Model id {model_id} is not supported. Please use one of {_MODEL_FN_DICT.keys()}')

        llm_proxy_model_id = model_id
        response = _MODEL_FN_DICT[model_id](req, llm_proxy_model_id, timeout=timeout, temperature=temperature)

        if response.status_code != 200:
            if response.status_code == 429:
                raise LLMProxyRateLimitError(f'Request failed with status code {response.status_code}: {response.text}')
            raise RuntimeError(f'Request failed with status code {response.status_code}: {response.text}')

        out = json.loads(response.json()['completion'])['choices'][0]['message']['content']
        return out

    if not num_retries:
        return _fn()

    retry_policy = Retrying(stop=stop_after_attempt(num_retries + 1),
                            wait=wait_exponential(multiplier=RETRY_WAIT_EXPONENTIAL_MULTIPLIER,
                                                  min=RETRY_WAIT_EXPONENTIAL_MIN,
                                                  max=RETRY_WAIT_EXPONENTIAL_MAX),
                            retry=retry_if_exception(lambda e: isinstance(e, LLMProxyRateLimitError)))
    return retry_policy(_fn)


def generate_base_model_using_chat_completion_messages(messages: List[Dict[str, str]],
                                                       model_clz: Any,
                                                       model_id: Optional[str] = None,
                                                       num_retries: int = 0) -> BaseModel:
    """Generates a base model using chat completion messages using structured output.

    Args:
        messages (list): A list of dictionaries representing the chat messages.
        model_clz (Any): The Pydantic model class to use for the response.
        model_id_suffix (str, optional): The suffix to append to the model id. Defaults to None.
        num_retries (int, optional): The number of retries for the request. Defaults to 0.

    Returns:
        BaseModel: The generated Pydantic model.
    """
    if not OPENAI_AVAILABLE:
        raise ImportError("openai library is not available. Please install it to use this")

    req: Dict = {
        'messages': messages,
        'response_format': type_to_response_format_param(model_clz),
    }
    llm_proxy_model_id = model_id or 'gpt-4o-2024-08-06'
    res = get_llm_proxy_chat_completion_response(llm_proxy_model_id, req, num_retries=num_retries)
    return model_clz(**json.loads(res))



"""Evaluator class for Key Information Extraction (KIE) tasks."""

import json
import logging
import math
import re
import statistics
from typing import Any, Dict, Optional, Type, Union

import mlflow
import numpy as np
import pandas as pd
from databricks.agents.evals import metric  # pylint: disable=ungrouped-imports
from mlflow.evaluation import Assessment
from mlflow.models import EvaluationResult
from pydantic import BaseModel, ValidationError
from pydantic.fields import FieldInfo
from pyspark.sql.dataframe import DataFrame
from scipy.optimize import linear_sum_assignment

from databricks.kie.constants import MLFLOW_REGISTERED_MODEL_TAG
from databricks.kie.eval_utils import (PrimitiveTypes, correctness, exact_match, fuzzy_match, guideline_adherence,
                                       normalized_match)
from databricks.model_serving.utils import ModelSpecs
from databricks.model_training.api.utils import get_spark
from databricks.model_training.logging import temp_log_level

# Information on general metrics

OVERALL_SCORE = 'overall_score'
OVERALL_SCORE_RATIONALE = 'Overall score is the average of the individual field scores'
IS_SCHEMA_MATCH_RATIONALE = 'The schema of the response matches the expected schema.'
IS_JSON_PARSABLE_RATIONALE = 'The response is valid JSON.'
IS_NOT_SCHEMA_MATCH_RATIONALE = 'The schema of the response does not match the expected schema.'
IS_NOT_JSON_PARSABLE_RATIONALE = 'The response is not valid JSON.'

# Information on null metrics

TRUE_POSITIVE_NULL_RATE = 'true_positive_null_rate'
FALSE_POSITIVE_NULL_RATE = 'false_positive_null_rate'
FALSE_NEGATIVE_NULL_RATE = 'false_negative_null_rate'
NULL_FIELD_RATE_GT = "null_field_rate_gt"
NULL_FIELD_RATE_PRED = "null_field_rate_pred"

TRUE_POSITIVE_NULL_RATE_RATIONALE = ('The rate of correct null extractions is {rate} ({numerator}/{denominator}). ' +
                                     'Used to determine how much of nulls expected are correctly predicted. ' +
                                     'Calculated as (null-pred-null-gt / total null-gt)')
FALSE_POSITIVE_NULL_RATE_RATIONALE = (
    'The rate of false positive null extractions is {rate} ({numerator}/{denominator}). ' +
    'Used to determine if the model is predicting nulls that the ground truth claims to have valid extractions. ' +
    'Calculated as (null-pred-non-null-gt / total non-null-gt)')
FALSE_NEGATIVE_NULL_RATE_RATIONALE = (
    'The rate of false negative null extractions is {rate} ({numerator}/{denominator}). ' +
    'Used to determine if the model is predicting values that the ground truth claims to be null. ' +
    'Calculated as (non-null-pred-null-gt / total null-gt)')
NULL_FIELD_RATE_GT_RATIONALE = ('The rate of null fields in the ground truth is {rate} ({numerator}/{denominator}). ' +
                                'Used to determine what % of the extractions are null in the ground truth' +
                                'Calculated as (null-gt / null-gt + non-null-gt)')
NULL_FIELD_RATE_PRED_RATIONALE = ('The rate of null fields in the prediction is {rate} ({numerator}/{denominator}). ' +
                                  'Used to determine what % of the extractions are null in the prediction' +
                                  'Calculated as (null-pred / null-pred + non-null-pred)')

NULL_INFO_RESULTS_KEYS = [
    TRUE_POSITIVE_NULL_RATE, FALSE_POSITIVE_NULL_RATE, FALSE_NEGATIVE_NULL_RATE, NULL_FIELD_RATE_GT,
    NULL_FIELD_RATE_PRED
]

# Null metric helper keys

NULL_INFO_REQUIREMENTS_KEYS = [
    "total_nulls_gt", "total_non_nulls_gt", "null_pred_non_null_gt", "non_null_pred_null_gt", "null_pred_null_gt",
    "total_nulls_pred", "total_non_nulls_pred"
]

# TODO figure out how enums come in
AllowedInnerTypes = Union[PrimitiveTypes, BaseModel, list['AllowedInnerTypes']]
FieldTuple = tuple[AllowedInnerTypes, FieldInfo]

JUDGE_NAME_GUIDELINE = 'guideline_judge'
KIE_EVAL_METRIC_REGISTRY = {
    'exact_match': exact_match,
    'normalized_match': normalized_match,
    'fuzzy_match': fuzzy_match,
    'correctness': correctness,
    JUDGE_NAME_GUIDELINE: guideline_adherence,
}
DISALLOWED_NESTED_METRICS = {'correctness', JUDGE_NAME_GUIDELINE}


def get_default_metric_from_field(field_info: FieldInfo, gt_value: PrimitiveTypes, from_root: bool) -> str:
    """Select a default metric to use based on the field information

    Note: In the future, this routing could be done by an LLM, or other more sophisticated method.

    Args:
        field_info (FieldInfo): Pydantic field information.
        gt_value (PrimitiveTypes): Ground truth value.
        from_root (bool): Whether the field is a root field.

    Returns:
        str: Default metric to use.
    """

    del field_info  # Currently unused, but might be used for the description in the future

    if not isinstance(gt_value, str):
        return 'exact_match'

    gt_length = len(gt_value)

    # Completely arbitrary threshold
    if gt_length <= 10:
        return 'normalized_match'
    elif not from_root or gt_length <= 50:
        return 'fuzzy_match'
    else:
        return JUDGE_NAME_GUIDELINE


def parse_json_markdown(text: str) -> str:
    """Parse JSON from a markdown-formatted string.

    Allows text before or after the JSON block, and if no
    JSON block is found, returns the entire text.

    Args:
        text (str): Markdown-formatted string.

    Returns:
        str: JSON string.
    """
    pattern = r'^.*?```(?:json)?\s*(.*?)```\s*.*$'
    match = re.search(pattern, text, re.DOTALL)
    if match:
        return match.group(1).strip()
    return text.strip()


def is_null_like(value: Any) -> bool:
    """Check if a value is null-like.

    Checks None, empty dict, empty list, and various string representations of null

    Args:
        value (any): Value to check.

    Returns:
        bool: True if value is null-like, False otherwise.
    """
    null_values = [None, '', 'null', 'none', 'nan', 'n/a', 'na', [], {}]
    is_none = value is None
    is_null_value = value in null_values
    lower_is_null_value = isinstance(value, str) and value.lower() in null_values
    is_nan = isinstance(value, float) and math.isnan(value)
    return is_none or is_null_value or lower_is_null_value or is_nan


class KIEEvaluator:
    """Evaluator for text to json output."""

    def __init__(self, schema: Type[BaseModel]):
        self.schema = schema

    def _create_null_information_assessments(self, null_info_requirements: dict[str, int]) -> dict[str, Assessment]:
        """Create null information assessments for a row."""

        # Helper function to format null metrics
        def _format_null_metrics(rate: float) -> float:
            return round(rate, 2)

        # Set the default of null metrics to None
        true_positive_null_assessment = Assessment(name=TRUE_POSITIVE_NULL_RATE,
                                                   value=None,
                                                   error_code="No nulls in ground truth")
        false_positive_null_assessment = Assessment(name=FALSE_POSITIVE_NULL_RATE,
                                                    value=None,
                                                    error_code="No nulls in ground truth")
        false_negative_null_assessment = Assessment(name=FALSE_NEGATIVE_NULL_RATE,
                                                    value=None,
                                                    error_code="No non-nulls in ground truth")
        null_field_rate_gt_assessment = Assessment(name=NULL_FIELD_RATE_GT,
                                                   value=None,
                                                   error_code="No nulls in ground truth")
        null_field_rate_pred_assessment = Assessment(name=NULL_FIELD_RATE_PRED,
                                                     value=None,
                                                     error_code="No nulls in prediction")
        # Calculate the necessary null metrics if there are any nulls in the ground truth
        if null_info_requirements["total_nulls_gt"] > 0:
            true_positive_null_rate = _format_null_metrics(null_info_requirements["null_pred_null_gt"] /
                                                           null_info_requirements["total_nulls_gt"])
            true_positive_null_assessment = Assessment(name=TRUE_POSITIVE_NULL_RATE,
                                                       value=true_positive_null_rate,
                                                       rationale=TRUE_POSITIVE_NULL_RATE_RATIONALE.format(
                                                           rate=true_positive_null_rate,
                                                           numerator=null_info_requirements["null_pred_null_gt"],
                                                           denominator=null_info_requirements["total_nulls_gt"]))
            false_negative_null_rate = _format_null_metrics(null_info_requirements["non_null_pred_null_gt"] /
                                                            null_info_requirements["total_nulls_gt"])
            false_negative_null_assessment = Assessment(name=FALSE_NEGATIVE_NULL_RATE,
                                                        value=false_negative_null_rate,
                                                        rationale=FALSE_NEGATIVE_NULL_RATE_RATIONALE.format(
                                                            rate=false_negative_null_rate,
                                                            numerator=null_info_requirements["non_null_pred_null_gt"],
                                                            denominator=null_info_requirements["total_nulls_gt"]))

        # Calculate the necessary null metrics if there are any non-nulls in the ground truth
        if null_info_requirements["total_non_nulls_gt"] > 0:
            false_positive_null_rate = _format_null_metrics(null_info_requirements["null_pred_non_null_gt"] /
                                                            null_info_requirements["total_non_nulls_gt"])
            false_positive_null_assessment = Assessment(name=FALSE_POSITIVE_NULL_RATE,
                                                        value=false_positive_null_rate,
                                                        rationale=FALSE_POSITIVE_NULL_RATE_RATIONALE.format(
                                                            rate=false_positive_null_rate,
                                                            numerator=null_info_requirements["null_pred_non_null_gt"],
                                                            denominator=null_info_requirements["total_non_nulls_gt"]))
        # Calculate the null field rate for the prediction
        total_pred_fields = null_info_requirements["total_nulls_pred"] + null_info_requirements["total_non_nulls_pred"]
        if total_pred_fields > 0:
            null_field_rate = _format_null_metrics(null_info_requirements["total_nulls_pred"] / total_pred_fields)
            null_field_rate_pred_assessment = Assessment(name=NULL_FIELD_RATE_PRED,
                                                         value=null_field_rate,
                                                         rationale=NULL_FIELD_RATE_PRED_RATIONALE.format(
                                                             rate=null_field_rate,
                                                             numerator=null_info_requirements["total_nulls_pred"],
                                                             denominator=total_pred_fields))
        # Calculate the null field rate for the ground truth
        total_gt_fields = null_info_requirements["total_nulls_gt"] + null_info_requirements["total_non_nulls_gt"]
        if total_gt_fields > 0:
            null_field_rate = _format_null_metrics(null_info_requirements["total_nulls_gt"] / total_gt_fields)
            null_field_rate_gt_assessment = Assessment(name=NULL_FIELD_RATE_GT,
                                                       value=null_field_rate,
                                                       rationale=NULL_FIELD_RATE_GT_RATIONALE.format(
                                                           rate=null_field_rate,
                                                           numerator=null_info_requirements["total_nulls_gt"],
                                                           denominator=total_gt_fields))
        return {
            TRUE_POSITIVE_NULL_RATE: true_positive_null_assessment,
            FALSE_POSITIVE_NULL_RATE: false_positive_null_assessment,
            FALSE_NEGATIVE_NULL_RATE: false_negative_null_assessment,
            NULL_FIELD_RATE_GT: null_field_rate_gt_assessment,
            NULL_FIELD_RATE_PRED: null_field_rate_pred_assessment
        }

    def evaluate_row(self, query: str, ground_truth_answer: str,
                     generated_answer: str) -> tuple[dict[str, Any], dict[str, Any]]:
        top_level_keys = self.schema.model_fields.keys()

        # Parse the provided strings into JSON strings
        gt_json_parsed = parse_json_markdown(ground_truth_answer)
        pred_json_parsed = parse_json_markdown(generated_answer)

        # Ground truth must be valid JSON
        try:
            gt_json_loaded = json.loads(gt_json_parsed)
        except json.JSONDecodeError as e:
            raise ValueError(f'Ground truth answer is not a valid JSON: {ground_truth_answer}') from e

        # Ground truth must match the provided schema
        try:
            gt_loaded = self.schema(**gt_json_loaded)
        except ValidationError as e:
            raise ValueError(f'Ground truth answer does not match the provided schema: {ground_truth_answer}') from e

        # Prediction may not have been valid JSON. Return early if it is not.
        try:
            pred_json_loaded = json.loads(pred_json_parsed)
        except json.JSONDecodeError as e:
            return {
                k: Assessment(name=k, value=0.0, rationale=IS_NOT_JSON_PARSABLE_RATIONALE) for k in top_level_keys
            }, {
                OVERALL_SCORE:
                    Assessment(name=OVERALL_SCORE, value=0.0, rationale=IS_NOT_JSON_PARSABLE_RATIONALE),
                'is_schema_match':
                    Assessment(name='is_schema_match', value=False, rationale=IS_NOT_JSON_PARSABLE_RATIONALE),
                'is_json_parsable':
                    Assessment(name='is_json_parsable', value=False, rationale=IS_NOT_JSON_PARSABLE_RATIONALE)
            }

        # Prediction may not match the provided schema. Return early if it is not.
        try:
            pred_loaded = self.schema(**pred_json_loaded)
        except ValidationError as e:
            return {
                k: Assessment(name=k, value=0.0, rationale=IS_NOT_SCHEMA_MATCH_RATIONALE) for k in top_level_keys
            }, {
                OVERALL_SCORE:
                    Assessment(name=OVERALL_SCORE, value=0.0, rationale=IS_NOT_SCHEMA_MATCH_RATIONALE),
                'is_schema_match':
                    Assessment(name='is_schema_match', value=False, rationale=IS_NOT_SCHEMA_MATCH_RATIONALE),
                'is_json_parsable':
                    Assessment(name='is_json_parsable', value=True, rationale=IS_NOT_SCHEMA_MATCH_RATIONALE)
            }

        # Both ground truth and prediction have been loaded into the provided schema and are pydantic objects.
        # TODO: Add an assert that the structures match at this point since all fields are required

        scored_output, null_info_requirements = self._scoring_helper_object(query, gt_loaded, pred_loaded, is_root=True)
        null_information = self._create_null_information_assessments(null_info_requirements)

        overall_score = statistics.mean([float(v.value) for v in scored_output.values()])
        overall_scoring = {
            OVERALL_SCORE: Assessment(name=OVERALL_SCORE, value=overall_score, rationale=OVERALL_SCORE),
            'is_schema_match': Assessment(name='is_schema_match', value=True, rationale=IS_SCHEMA_MATCH_RATIONALE),
            'is_json_parsable': Assessment(name='is_json_parsable', value=True, rationale=IS_JSON_PARSABLE_RATIONALE),
            **null_information
        }
        return scored_output, overall_scoring

    def _compute_pairwise_scores(
        self,
        query: str,
        gt_items: list[AllowedInnerTypes],
        pred_items: list[AllowedInnerTypes],
        gt_field_info: FieldInfo,
    ) -> np.ndarray:
        """Compute pairwise similarity scores between ground truth and predicted items."""
        scores = np.zeros((len(gt_items), len(pred_items)))
        null_info_matrix = [[None for _ in range(len(pred_items))] for _ in range(len(gt_items))]

        for i, gt_item in enumerate(gt_items):
            for j, pred_item in enumerate(pred_items):
                score_value, null_info_entry = self._scoring_helper_recurse(query, (gt_item, gt_field_info),
                                                                            (pred_item, gt_field_info))
                scores[i, j] = score_value.value
                null_info_matrix[i][j] = null_info_entry

        return scores, null_info_matrix

    def _compute_assignment(self, scores: np.ndarray) -> tuple[list[int], list[int]]:
        """Find optimal assignment using Hungarian algorithm."""
        row_ind, col_ind = linear_sum_assignment(scores, maximize=True)
        return row_ind.tolist(), col_ind.tolist()

    def _calculate_null_information_from_list(self, assignments,
                                              null_info_matrix: list[list[dict[str, int]]]) -> dict[str, int]:
        """Calculate null information from a list of items."""
        row_ind, col_ind = assignments
        matched_null_info = []
        for gt_idx, pred_idx in zip(row_ind, col_ind):
            matched_null_info.append(null_info_matrix[gt_idx][pred_idx])

        # Calculate each of the null information metrics for the entire list
        null_information = {k: sum(info[k] for info in matched_null_info) for k in NULL_INFO_REQUIREMENTS_KEYS}
        return null_information

    def _compute_precision_recall_f1(self, scores: np.ndarray,
                                     assignments: tuple[list[int], list[int]]) -> tuple[float, float, float]:
        """Compute precision and recall based on assignments."""
        row_ind, col_ind = assignments
        sum_scores = scores[row_ind, col_ind].sum()

        num_gt_items = scores.shape[0]
        num_pred_items = scores.shape[1]

        precision = sum_scores / num_pred_items
        recall = sum_scores / num_gt_items
        f1 = 2 * precision * recall / (precision + recall) if (precision + recall) > 0 else 0

        return precision, recall, f1

    def _scoring_helper_object(self,
                               query: str,
                               ground_truth_answer: BaseModel,
                               generated_answer: BaseModel,
                               is_root: bool = False) -> dict[str, Assessment]:
        results = {}
        null_information = {k: 0 for k in NULL_INFO_REQUIREMENTS_KEYS}
        for key_name in ground_truth_answer.model_dump():
            gt_value = getattr(ground_truth_answer, key_name)
            pred_value = getattr(generated_answer, key_name)

            gt_field_info = ground_truth_answer.model_fields[key_name]
            pred_field_info = generated_answer.model_fields[key_name]

            score, null_info_entry = self._scoring_helper_recurse(query, (gt_value, gt_field_info),
                                                                  (pred_value, pred_field_info),
                                                                  field_name=key_name,
                                                                  from_root=is_root)
            results[key_name] = Assessment(
                name=f'{key_name}',
                value=score.value,
                rationale=score.rationale,
            )
            for k in NULL_INFO_REQUIREMENTS_KEYS:
                null_information[k] += null_info_entry[k]

        return results, null_information

    def _score_null_presence(self, gt_value: PrimitiveTypes,
                             pred_value: PrimitiveTypes) -> tuple[Assessment, Dict[str, int]]:
        null_information = {k: 0 for k in NULL_INFO_REQUIREMENTS_KEYS}

        gt_is_null_like = is_null_like(gt_value)
        pred_is_null_like = is_null_like(pred_value)
        assert gt_is_null_like or pred_is_null_like, f'Unsupported null presence case: {gt_value} and {pred_value}'

        # Case 1: Both values are null-like
        if gt_is_null_like and pred_is_null_like:
            null_information["total_nulls_gt"] = 1
            null_information["total_nulls_pred"] = 1
            null_information["null_pred_null_gt"] = 1
            return Assessment(name='null_match', value=True, rationale='Both values are null-like'), null_information

        # Case 2: Ground truth is null-like but prediction is not
        if gt_is_null_like:
            null_information["total_nulls_gt"] = 1
            null_information["total_non_nulls_pred"] = 1
            null_information["non_null_pred_null_gt"] = 1
            return Assessment(name='null_match',
                              value=False,
                              rationale='expected_response is null-like but response is not'), null_information

        # Case 3: Prediction is null-like but ground truth is not
        if pred_is_null_like:
            null_information["total_non_nulls_gt"] = 1
            null_information["total_nulls_pred"] = 1
            null_information["null_pred_non_null_gt"] = 1
            return Assessment(name='null_match',
                              value=False,
                              rationale='response is null-like but expected_response is not'), null_information

        # Raise ValueError here because this should never happen
        raise ValueError(f'Unsupported null presence case: {gt_value} and {pred_value}')

    def _scoring_helper_recurse(self,
                                query: str,
                                ground_truth_answer: FieldTuple,
                                generated_answer: FieldTuple,
                                field_name: Optional[str] = None,
                                from_root: bool = False) -> tuple[Assessment, Dict[str, int]]:
        gt_value, gt_field_info = ground_truth_answer
        null_information = {k: 0 for k in NULL_INFO_REQUIREMENTS_KEYS}

        pred_value = generated_answer[0]

        # Early exit if either value is null-like
        gt_is_null_like = is_null_like(gt_value)
        pred_is_null_like = is_null_like(pred_value)
        if gt_is_null_like or pred_is_null_like:
            assessment, null_information = self._score_null_presence(gt_value, pred_value)
            return assessment, null_information

        if isinstance(gt_value, PrimitiveTypes):
            assert isinstance(pred_value, PrimitiveTypes)

            default_kie_eval_metric = get_default_metric_from_field(gt_field_info, gt_value, from_root)

            if gt_field_info.json_schema_extra is not None:
                eval_metric_string = str(
                    gt_field_info.json_schema_extra.get(  # type: ignore
                        'eval_metric', default_kie_eval_metric))
                eval_metric_kwargs: dict = gt_field_info.json_schema_extra.get('eval_metric_kwargs', {})  # type: ignore
            else:
                eval_metric_string = default_kie_eval_metric
                eval_metric_kwargs = {}

            if eval_metric_string == JUDGE_NAME_GUIDELINE:
                eval_metric_kwargs['field_name'] = field_name
                eval_metric_kwargs['field_definition'] = gt_field_info.description

            if eval_metric_string not in KIE_EVAL_METRIC_REGISTRY:
                raise ValueError(f'Unsupported evaluation metric: {eval_metric_string}')

            eval_metric_func = KIE_EVAL_METRIC_REGISTRY[eval_metric_string]

            assert isinstance(eval_metric_kwargs, dict)

            if eval_metric_string in DISALLOWED_NESTED_METRICS and not from_root:
                raise ValueError(f'Unsupported evaluation metric {eval_metric_string} for nested field')

            score = eval_metric_func(query, gt_value, pred_value, **eval_metric_kwargs)

            null_information["total_non_nulls_gt"] = 1
            null_information["total_non_nulls_pred"] = 1
            return score, null_information
        elif isinstance(gt_value, BaseModel):
            assert isinstance(pred_value, BaseModel)
            per_key_results, null_information = self._scoring_helper_object(query, gt_value, pred_value)
            result = statistics.mean([float(v.value) for v in per_key_results.values()])
            return Assessment(name='object_score', value=result,
                              rationale='Average of the individual field scores'), null_information
        elif isinstance(gt_value, list):
            assert isinstance(pred_value, list)
            pairwise_scores, null_info_matrix = self._compute_pairwise_scores(query, gt_value, pred_value,
                                                                              gt_field_info)
            assignments = self._compute_assignment(pairwise_scores)
            _, _, f1 = self._compute_precision_recall_f1(pairwise_scores, assignments)
            null_information = self._calculate_null_information_from_list(assignments, null_info_matrix)
            return Assessment(
                name='list_score',
                value=f1,
                rationale='Soft F1 score of the optimal pairwise assignment between list items'), null_information
        else:
            raise ValueError(f'Unsupported type: {type(gt_value)}')


def store_evaluation_metrics(df: pd.DataFrame, response_format: Type[BaseModel]) -> EvaluationResult:
    eval_class = KIEEvaluator(response_format)

    @metric
    def kie(request, response, expected_response):
        eval_output_per_key, eval_output_overall = eval_class.evaluate_row(request, expected_response, response)
        all_assessments = [eval_output_overall['is_schema_match']] + list(
            eval_output_per_key.values()) + [eval_output_overall[OVERALL_SCORE]]
        null_info_results_keys = [k for k in NULL_INFO_RESULTS_KEYS if k in eval_output_overall]
        all_assessments.extend(eval_output_overall[k] for k in null_info_results_keys)
        return all_assessments

    eval_result = mlflow.evaluate(
        model=None,
        data=df,
        model_type='databricks-agent',
        evaluator_config={'databricks-agent': {
            'metrics': [],
        }},
        extra_metrics=[kie],
    )

    return eval_result


def evaluate_model(df: DataFrame,
                   experiment_id: str,
                   model: ModelSpecs,
                   response_format: type[BaseModel],
                   run_name: Optional[str] = None,
                   run_id: Optional[str] = None) -> DataFrame:

    for required_column in ('request', 'response', 'expected_response'):
        if required_column not in df.columns:
            raise ValueError(f'`{required_column}` column is required for evaluation')

    model_df = df.toPandas()

    # Log each metric to mlflow
    start_kwargs = {}
    if run_id:
        start_kwargs['run_id'] = run_id
    elif run_name:
        start_kwargs['run_name'] = run_name
    with mlflow.start_run(experiment_id=experiment_id, **start_kwargs), temp_log_level('mlflow', logging.ERROR):
        mlflow.log_param(key='cost', value=model.cost_per_hour)
        mlflow.set_tag(MLFLOW_REGISTERED_MODEL_TAG, model.uc_schema)
        eval_result = store_evaluation_metrics(model_df, response_format)
        result_df = eval_result.tables['eval_results']

    spark = get_spark()
    return spark.createDataFrame(result_df)



"""Script used to upload the T2T system parameter analyzer model to model registry."""

import argparse
import logging
import os
import shutil

import mlflow
from databricks.sdk.core import Config
from mlflow.models.signature import ModelSignature, ParamSchema
from mlflow.types.schema import ColSpec, ParamSpec, Schema
from t2t_system_param_analyzer_model import T2TSystemParamAnalyzerModel

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

mlflow.set_tracking_uri("databricks")
mlflow.set_registry_uri("databricks-uc")


def save_model(local_output_path="t2t_system_param_analyzer_model"):
    """Save the T2T system parameter analyzer model with artifacts."""
    model = T2TSystemParamAnalyzerModel()

    input_schema = Schema([
        ColSpec("string", "request"),
        ColSpec("string", "rejected_output", required=False),
        ColSpec("string", "preferred_output", required=False)
    ])
    # TODO(jun.choi): Correctly specify schema as nested dictionary.
    output_schema = Schema([
        ColSpec("string", "new_task_spec"),
    ])
    params_schema = ParamSchema([
        ParamSpec("current_task_spec", "string", None),
    ])
    signature = ModelSignature(inputs=input_schema, outputs=output_schema, params=params_schema)

    current_dir = os.path.dirname(os.path.abspath(__file__))
    t2t_system_param_analyzer_model_path = os.path.join(current_dir, "t2t_system_param_analyzer_model.py")

    mlflow.pyfunc.save_model(path=local_output_path,
                             python_model=model,
                             signature=signature,
                             code_paths=[t2t_system_param_analyzer_model_path],
                             pip_requirements=[
                                 'databricks-connect>=13.0',
                                 'databricks-genai==1.2.0a19',
                             ])


def upload(model_registry_path: str,
           local_output_path: str = "t2t_system_param_analyzer_model",
           profile: str = "DEFAULT"):
    """Upload the T2T system parameter analyzer model to the staging registry.
    Args:
        model_registry_path: Path in the model registry to upload the model to
        local_output_path: Local path to save the model artifacts
        profile: Profile to use from ~/.databrickscfg
    """
    config = Config(profile=profile)
    os.environ["DATABRICKS_HOST"] = config.host
    os.environ["DATABRICKS_TOKEN"] = config.token

    try:
        save_model(local_output_path)

        logger.info(f"Registering system parameter analyzer model to the workspace {config.host}"
                    f" at model registry path {model_registry_path}")
        mlflow.register_model(model_uri=local_output_path, name=model_registry_path)
        logger.info(f"Registered model to {model_registry_path}")

    except (ImportError, mlflow.exceptions.MlflowException) as e:
        logger.error(f"Failed to register model: {e}")
    finally:
        if os.path.exists(local_output_path) and os.path.isdir(local_output_path):
            shutil.rmtree(local_output_path)
            logger.info(f"Deleted existing model at {local_output_path}")


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Upload T2T system parameter analyzer model")
    parser.add_argument("--model_registry_path",
                        type=str,
                        required=True,
                        help="Path in the model registry to upload the system parameter analyzer model to")
    parser.add_argument("--profile", type=str, default="DEFAULT", help="Profile to use from ~/.databrickscfg")
    parser.add_argument("--local_output_path",
                        default="t2t_system_param_analyzer_model",
                        type=str,
                        help="Path to save the model locally")
    args = parser.parse_args()

    upload(model_registry_path=args.model_registry_path, local_output_path=args.local_output_path, profile=args.profile)



"""Script used to upload the KIE grounding model to model registry."""

import argparse
import logging
import os
import shutil

import mlflow
from databricks.sdk.core import Config
from mlflow.models.signature import ModelSignature, ParamSchema
from mlflow.types.schema import ColSpec, ParamSpec, Schema
from t2t_grounding_model import T2TGroundingModel

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

mlflow.set_tracking_uri("databricks")
mlflow.set_registry_uri("databricks-uc")


def save_model(local_output_path="grounding_model"):
    """Save the T2T grounding model with artifacts."""
    model = T2TGroundingModel()

    input_schema = Schema([ColSpec("string", "request"), ColSpec("string", "expected_response", required=False)])
    output_schema = Schema([
        ColSpec("string", "request"),
        ColSpec("string", "expected_response", required=False),
        ColSpec("string", "generated_response"),
        ColSpec("string", "eval_results"),
    ])
    params_schema = ParamSchema([ParamSpec("system_parameter", "string", None)])
    signature = ModelSignature(inputs=input_schema, outputs=output_schema, params=params_schema)

    current_dir = os.path.dirname(os.path.abspath(__file__))
    t2t_grounding_model_path = os.path.join(current_dir, "t2t_grounding_model.py")

    mlflow.pyfunc.save_model(path=local_output_path,
                             python_model=model,
                             signature=signature,
                             code_paths=[t2t_grounding_model_path],
                             pip_requirements=[
                                 'pandas>=1.5.0',
                                 'requests>=2.28.0',
                                 'mlflow>=2.8.0',
                                 'databricks-connect>=13.0',
                                 'pydantic>=2.0.0',
                                 'databricks-genai==1.2.0a20',
                             ])


def upload(model_registry_path: str, local_output_path: str = "grounding_model", profile: str = "DEFAULT"):
    """Upload the T2T grounding model to the staging registry.

    Args:
        model_registry_path: Path in the model registry to upload the model to
        local_output_path: Local path to save the model artifacts
        profile: Profile to use from ~/.databrickscfg
    """
    config = Config(profile=profile)
    os.environ["DATABRICKS_HOST"] = config.host
    os.environ["DATABRICKS_TOKEN"] = config.token

    try:
        save_model(local_output_path)

        logger.info(
            f"Registering grounding model to the workspace {config.host} at model registry path {model_registry_path}")
        mlflow.register_model(model_uri=local_output_path, name=model_registry_path)
        logger.info(f"Registered model to {model_registry_path}")

    except (ImportError, mlflow.exceptions.MlflowException) as e:
        logger.error(f"Failed to register model: {e}")
    finally:
        if os.path.exists(local_output_path) and os.path.isdir(local_output_path):
            shutil.rmtree(local_output_path)
            logger.info(f"Deleted existing model at {local_output_path}")


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Upload T2T grounding model")
    parser.add_argument("--model_registry_path",
                        type=str,
                        required=True,
                        help="Path in the model registry to upload the grounding model to")
    parser.add_argument("--profile", type=str, default="DEFAULT", help="Profile to use from ~/.databrickscfg")
    parser.add_argument("--local_output_path",
                        default="grounding_model",
                        type=str,
                        help="Path to save the model locally")
    args = parser.parse_args()

    upload(model_registry_path=args.model_registry_path, local_output_path=args.local_output_path, profile=args.profile)



"""Pyfunc model abstraction for T2T grounding."""

import json
import logging
import re
from concurrent.futures import ThreadPoolExecutor, as_completed
from typing import Any, Dict, List

import pandas as pd
from mlflow.pyfunc import PythonModel

from databricks.kie.eval_utils import guideline_adherence_with_evaluation_criteria
from databricks.kie.t2t_models.base_t2t_model import BaseT2TModel
from databricks.kie.t2t_models.t2t_model_registrations import get_t2t_default_model
from databricks.kie.t2t_schema import T2TSystemParams

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


class T2TGroundingModel(PythonModel):
    """MLflow model for T2T grounding."""
    MAX_WORKERS = 3

    def load_context(self, context) -> None:
        """Load the model artifacts from the context."""
        # pylint: disable=import-outside-toplevel
        import mlflow
        mlflow.set_tracking_uri("databricks")
        mlflow.set_registry_uri("databricks-uc")

    def evaluate(self, df: pd.DataFrame, system_params: T2TSystemParams) -> pd.DataFrame:
        """Generate evaluation results."""
        evaluation_criteria = system_params.evaluation_criteria

        def process_row(row) -> Dict[str, Any]:
            """Process a single row ensuring all inputs remain as strings."""
            request = str(row['request'])
            response_str = row['generated_response']
            if not response_str:
                return {}

            result = {}
            for criterion in evaluation_criteria:
                metric_name = re.sub(r'[^a-zA-Z0-9]', '_', criterion)
                assessment = guideline_adherence_with_evaluation_criteria(request, response_str, criterion)

                result[f"metric/t2t/{metric_name}/value"] = assessment.value
                result[f"metric/t2t/{metric_name}/rationale"] = assessment.rationale

            return result

        df_results = pd.DataFrame(df.apply(process_row, axis=1))
        logger.info(f'Finished generating evaluation results for {len(df_results)} rows')

        return df_results

    def _generate_responses(self, requests: List[str], model: BaseT2TModel) -> List[str]:
        """Generate responses for requests without expected responses.

        Args:
            requests: List of request strings to generate responses for
            model: T2T model to use for generating responses
        Returns:
            List of generated responses.
        """
        results = [None] * len(requests)
        with ThreadPoolExecutor(max_workers=5) as executor:
            # Map each future to its original index for ordering.
            future_to_index = {executor.submit(model, n): idx for idx, n in enumerate(requests)}

        # In order to prevent program crash when a single request fails,
        # we use `submit()` and instead catch all exceptions.
        # If request fails, we set the result to None and this will be skipped
        # during the grounding loop.
        for future in as_completed(future_to_index):
            idx = future_to_index[future]
            try:
                results[idx] = future.result()
            except RuntimeError as e:
                logger.warning(f"LLMProxyError: Error generating response for request {requests[idx]}: {e}")
                results[idx] = None

        return results

    def predict(self, context, model_input: pd.DataFrame, params: Dict[str, Any] = None):
        """Label T2T grounding samples and return evaluation results.

        Args:
            context: MLflow context
            model_input: DataFrame with `request` and optionally `expected_response` columns.
            params: Dictionary with `system_parameter` key

        Returns:
            DataFrame with generated_response and eval_results columns
        """
        del context

        if params is None or 'system_parameter' not in params:
            raise ValueError(
                "Expected params to be a dict containing the 'system_parameter' for grounding but found None.")

        system_params_str = params['system_parameter']
        try:
            system_params_json = json.loads(system_params_str)
        except json.JSONDecodeError as e:
            raise ValueError(f"Failed to parse 'system_parameter': {e}") from e

        system_params = T2TSystemParams(**system_params_json)
        model = get_t2t_default_model(system_params)
        df = model_input.copy()
        requests: List[str] = df['request'].tolist()
        df['generated_response'] = self._generate_responses(requests, model)
        df['eval_results'] = self.evaluate(df, system_params)

        return df



"""Iniitialize the KIE task and load cached values"""
import functools
import json
import os
from dataclasses import dataclass
from enum import IntEnum
from pathlib import Path
from typing import Optional, Tuple, Type
from urllib.parse import urljoin

import mlflow
import pandas as pd
from databricks.sdk import WorkspaceClient  # pylint: disable = ungrouped-imports
from mlflow.entities import Experiment
from pydantic import BaseModel
from pyspark.sql import DataFrame
from pyspark.sql.types import StringType, StructField, StructType

from databricks.kie.data_utils import cache_df, split_labeled_data, split_unlabeled_data
from databricks.kie.kie_schema import ModelFactory
from databricks.kie.kie_state import KIEState
from databricks.kie.prompt_builder import PromptBuilder
from databricks.kie.task_spec import KIETaskSpec
from databricks.model_training.api.utils import get_browser_url, get_schema_from_table, get_spark

_SHARE_KEY = "kie_view"


class CacheLevel(IntEnum):
    """Amount of data to cache"""
    # No caching at all
    NONE = 1
    # Cache only the JSON schema
    JSON_SCHEMA = 2
    # Cache JSON schema and intermediate tables
    SCHEMA_AND_DF = 3

    @classmethod
    def from_str(cls, level: str) -> 'CacheLevel':
        try:
            return cls[level.upper()]
        except KeyError as e:
            raise ValueError(f"Invalid CacheLevel: {level}. Must be one of {', '.join(cls.__members__)}") from e

    def includes(self, other: 'CacheLevel') -> bool:
        return self >= other


def get_table_if_exists(path: str) -> Optional[DataFrame]:
    # Load files from cache if they exist
    wc = WorkspaceClient()
    try:
        if wc.tables.exists(path).table_exists:
            spark = get_spark()
            return spark.read.table(path)
    except:  # pylint: disable=bare-except
        return None


def get_mlflow_experiment_link(experiment_id: str) -> str:
    """
    Get a link for the MLflow experiment using the given link text.
    This link will open the MLflow experiment page in a new tab.
    """
    host = get_browser_url()
    path = urljoin(host, f'/ml/experiments/{experiment_id}?viewStateShareKey={_SHARE_KEY}')
    return path


@dataclass
class CachePaths:
    """Holds all cache paths for a KIE task."""
    unlabeled_table: str
    labeled_table: str
    grounding_table: str
    val_table: str
    train_jsonl: str
    train_table: Optional[str]
    schema_json: str

    @classmethod
    def from_task_spec(cls, task_spec: KIETaskSpec, root_name: str) -> 'CachePaths':
        output_schema = task_spec.output_schema or get_schema_from_table(task_spec.output_table)
        cache_root = f"{output_schema}.{root_name}"

        return cls(unlabeled_table=f"{cache_root}_unlabeled",
                   labeled_table=f"{cache_root}_labeled",
                   grounding_table=f"{cache_root}_grounding",
                   val_table=f"{cache_root}_val",
                   train_jsonl=str(Path(task_spec.output_path) / f"{root_name}_train.jsonl"),
                   train_table=f"{cache_root}_train" if task_spec.output_schema else None,
                   schema_json=str(Path(task_spec.output_path) / f"{root_name}_schema.json"))


class TaskInitializer:
    """Initializes a KIE task's state from a task spec"""

    def __init__(self, task_spec: KIETaskSpec, cache_level: CacheLevel = CacheLevel.SCHEMA_AND_DF):
        self.task_spec = task_spec
        self.spark = get_spark()
        self.cache_level = cache_level

    def _setup_experiment(self) -> Experiment:
        """Setup and return MLflow experiment."""
        experiment = mlflow.get_experiment_by_name(self.task_spec.experiment_name)
        if experiment is None:
            experiment_id = mlflow.create_experiment(self.task_spec.experiment_name)
            experiment = mlflow.get_experiment(experiment_id)
            path = get_mlflow_experiment_link(experiment.experiment_id)
            print(f"🧪 Created a new MLflow experiment at: {path}")
        else:
            path = get_mlflow_experiment_link(experiment.experiment_id)
            print(f"🧪 Using MLflow experiment at: {path}")

        mlflow.set_experiment(experiment_id=experiment.experiment_id)
        return experiment

    def _get_response_format(self, cache_paths: CachePaths) -> Type[BaseModel]:
        """Get or create response format."""
        if self.cache_level.includes(CacheLevel.JSON_SCHEMA) and os.path.exists(cache_paths.schema_json):
            return ModelFactory.from_file(cache_paths.schema_json)

        if self.task_spec.labeled_dataset and not self.task_spec.json_examples:
            labeled_df = self.spark.read.table(self.task_spec.labeled_dataset)
            rows = labeled_df.limit(10).collect()
            self.task_spec.json_examples = [
                json.loads(r[self.task_spec.labeled_dataset_output_json_column]) for r in rows
            ]

        return ModelFactory.from_examples(self.task_spec.json_examples, "response_format")

    def _setup_data_splits(self, cache_paths: CachePaths, num_grounding_samples: int,
                           num_val_samples: int) -> Tuple[Optional[DataFrame], DataFrame]:
        """Setup labeled and unlabeled data splits."""

        labeled_split_df = get_table_if_exists(
            cache_paths.labeled_table) if self.cache_level is CacheLevel.SCHEMA_AND_DF else None
        if self.task_spec.labeled_dataset and labeled_split_df is None:
            labeled_df = self.spark.read.table(self.task_spec.labeled_dataset)
            labeled_split_df = split_labeled_data(labeled_df,
                                                  num_grounding_samples=num_grounding_samples,
                                                  num_val_samples=num_val_samples)

            if self.cache_level is CacheLevel.SCHEMA_AND_DF:
                # Optionally cache the labeled dataset split
                cache_df(labeled_split_df, cache_paths.labeled_table)
            num_val_samples = num_grounding_samples = 0
            print(f"Sourcing labeled data from {self.task_spec.labeled_dataset}")

        unlabeled_split_df = get_table_if_exists(
            cache_paths.unlabeled_table) if self.cache_level is CacheLevel.SCHEMA_AND_DF else None
        if unlabeled_split_df is None:
            unlabeled_split_df = split_unlabeled_data(self.task_spec.unlabeled_dataset,
                                                      self.task_spec.unlabeled_delta_table,
                                                      self.task_spec.unlabeled_delta_table_text_column,
                                                      num_val_samples=num_val_samples,
                                                      num_grounding_samples=num_grounding_samples)

            if self.cache_level is CacheLevel.SCHEMA_AND_DF:
                # Optionally cache the unlabeled dataset split
                cache_df(unlabeled_split_df, cache_paths.unlabeled_table)

        print(f"Found {unlabeled_split_df.count()} unlabeled documents to extract")
        return labeled_split_df, unlabeled_split_df

    def validate_labeled_data_with_response_format(self, labeled_split_df: DataFrame, response_format: Type[BaseModel],
                                                   request_column: str, expected_response_column: str) -> None:
        """Validate labeled data against response format."""

        def _validate_pandas_df(iterator, response_format_json: str, request_column: str,
                                expected_response_column: str):
            # TODO: Fix this to not require serializing and deserializing the response_format
            # This will be brittle to changes in the response_format, especially as we move to
            # using the UI to define the json schema
            response_format = ModelFactory.from_json_schema(response_format_json)
            for pdf in iterator:
                failed_examples = []
                pdf = pd.DataFrame(pdf)
                for _, row in pdf.iterrows():
                    try:
                        response_format.model_validate(json.loads(row[expected_response_column]))
                    except Exception as e:  # pylint: disable=broad-except
                        failed_examples.append({
                            "request": row[request_column],
                            "expected_response": row[expected_response_column],
                            "error": str(e)
                        })
                result_df = pd.DataFrame(failed_examples)
                if result_df.empty:
                    result_df = pd.DataFrame(columns=["request", "expected_response", "error"])
                yield result_df

        validate_pandas_df = functools.partial(_validate_pandas_df,
                                               response_format_json=response_format.model_json_schema(),
                                               request_column=request_column,
                                               expected_response_column=expected_response_column)
        schema = StructType([
            StructField("request", StringType()),
            StructField("expected_response", StringType()),
            StructField("error", StringType())
        ])
        failed_examples = labeled_split_df.mapInPandas(validate_pandas_df, schema=schema)
        if failed_examples.count() > 0:
            error_msg = f"Found {failed_examples.count()} failed examples." + \
                "Please review the displayed data and try again."
            print(error_msg)
            failed_examples.display()
            raise ValueError(error_msg)
        else:
            print("✔ Validated labeled dataset's json output against expected response format.")

    def initialize(self,
                   num_grounding_samples: int,
                   num_val_samples: int,
                   num_fewshot_samples: int,
                   initialize_cache: bool = True) -> KIEState:
        """Initialize KIE task state."""
        experiment = self._setup_experiment()
        root_name = os.path.basename(experiment.name).replace("-", "_")

        if initialize_cache:
            # Create the output cache path if it doesn't exist
            os.makedirs(self.task_spec.output_path, exist_ok=True)

        cache_paths = CachePaths.from_task_spec(self.task_spec, root_name)

        response_format = self._get_response_format(cache_paths)
        prompt_builder = PromptBuilder(response_format)

        labeled_split_df, unlabeled_split_df = self._setup_data_splits(cache_paths, num_grounding_samples,
                                                                       num_val_samples)
        if self.task_spec.labeled_dataset:
            assert self.task_spec.labeled_dataset_text_column and self.task_spec.labeled_dataset_output_json_column
            self.validate_labeled_data_with_response_format(labeled_split_df, response_format,
                                                            self.task_spec.labeled_dataset_text_column,
                                                            self.task_spec.labeled_dataset_output_json_column)
        grounding_df = get_table_if_exists(
            cache_paths.grounding_table) if self.cache_level is CacheLevel.SCHEMA_AND_DF else None
        val_df = get_table_if_exists(cache_paths.val_table) if self.cache_level is CacheLevel.SCHEMA_AND_DF else None

        return KIEState(
            experiment=experiment,
            prompt_builder=prompt_builder,
            ground_truth_prompt=prompt_builder.build_prompt(),
            zeroshot_prompt=prompt_builder.build_prompt(include_markdown=True),
            fewshot_prompt=None,
            grounding_table_path=cache_paths.grounding_table,
            val_table_path=cache_paths.val_table,
            train_jsonl_path=cache_paths.train_jsonl,
            train_table_path=cache_paths.train_table,
            schema_path=cache_paths.schema_json,
            requires_grounding=grounding_df is None,
            requires_val=val_df is None,
            requires_train_gen=not os.path.exists(cache_paths.train_jsonl),
            unlabeled_split_df=unlabeled_split_df,
            labeled_split_df=labeled_split_df,
            grounding_df=grounding_df,
            val_df=val_df,
            response_format=response_format,
            num_grounding_samples=num_grounding_samples,
            num_val_samples=num_val_samples,
            num_fewshot_samples=num_fewshot_samples,
            model_dfs=None,
            unlabeled_table=cache_paths.unlabeled_table,
            labeled_table=cache_paths.labeled_table,
        )



"""Pyfunc model abstraction for T2T system parameter recommendation."""

import json
import logging
from dataclasses import is_dataclass
from typing import Any, Dict

import mlflow
import pandas as pd
from mlflow.pyfunc import PythonModel
from pydantic import BaseModel

from databricks.kie.t2t_evaluation_criteria_generator import (add_evaluation_criteria_from_preference_data,
                                                              generate_evaluation_criteria)
from databricks.kie.t2t_instruction_analyzer import populate_instruction_info
from databricks.kie.t2t_schema import PreferenceData, T2TSystemParams

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


# TODO(jun.choi): Include this as part of SDK t2t_utils.py instead.
def custom_asdict(obj):
    """Convert a dataclass or pydantic object to a dictionary."""
    if isinstance(obj, BaseModel):
        return obj.model_dump()
    elif isinstance(obj, list):
        return [custom_asdict(item) for item in obj]
    elif isinstance(obj, dict):
        return {k: custom_asdict(v) for k, v in obj.items()}
    elif is_dataclass(obj):
        return {field.name: custom_asdict(getattr(obj, field.name)) for field in obj.__dataclass_fields__.values()}
    else:
        return obj


class T2TSystemParamAnalyzerModel(PythonModel):
    """MLflow model for T2T system parameter analyzer."""

    def load_context(self, context) -> None:
        """Load the model artifacts from the context."""
        mlflow.set_tracking_uri("databricks")
        mlflow.set_registry_uri("databricks-uc")

    def predict(self, context, model_input: pd.DataFrame, params: Dict[str, Any] = None):
        """Label T2T grounding samples and return evaluation results.
        Args:
            context: MLflow context
            model_input: DataFrame with `request`, `rejected_output`, `preferred_output` columns.
            params: Dictionary with `current_task_spec` key.

        Returns:
            Dictionary containing details on a new T2T system parameters after grounding.
        """
        del context

        if params is None or 'current_task_spec' not in params:
            raise ValueError("Expected params to be a dict containing the 'current_task_spec' key.")

        # check that `model_input` has all required columns
        if not all(col in model_input.columns for col in ['request', 'rejected_output', 'preferred_output']):
            raise ValueError("model_input should have 'request', 'rejected_output', 'preferred_output' columns.")

        preference_data = []
        for request, rejected_output, preferred_output in model_input[[
                'request', 'rejected_output', 'preferred_output'
        ]].itertuples():
            preference_data.append(
                PreferenceData(input=request, rejected_response=rejected_output, preferred_response=preferred_output))

        current_task_spec_str = params['current_task_spec']
        try:
            task_spec_json = json.loads(current_task_spec_str)
        except json.JSONDecodeError as e:
            raise ValueError(f"Failed to parse 'current_task_spec': {e}") from e

        current_task_spec = T2TSystemParams(**task_spec_json)
        # TODO(jun.choi): Use diff between previous and current task spec.
        # previous_task_spec = T2TSystemParams(**params['previous_task_spec'])
        if not current_task_spec.evaluation_criteria:
            inferred_eval_criteria = generate_evaluation_criteria(current_task_spec.instruction,
                                                                  current_task_spec.labelled_test_examples)
            current_task_spec.evaluation_criteria = inferred_eval_criteria

        # If user provided labels or corrections, then add them to the evaluation criteria.
        if preference_data:
            new_eval_criteria = add_evaluation_criteria_from_preference_data(
                current_task_spec.instruction,
                current_task_spec.evaluation_criteria,
                preference_data,
            )
            current_task_spec.evaluation_criteria = new_eval_criteria

        populate_instruction_info(current_task_spec)
        return {"task_spec": custom_asdict(current_task_spec)}



"""Simple utils for working with KIE datasets"""
import json
import os
import random
from copy import deepcopy
from functools import wraps
from typing import Any, Dict, Iterable, List, Optional, Tuple, Type

import pandas as pd
from pydantic import BaseModel, ValidationError
from pyspark.sql import functions as F
from pyspark.sql.dataframe import DataFrame
from pyspark.sql.functions import monotonically_increasing_id, rand
from pyspark.sql.types import ArrayType, StringType, StructField, StructType
from pyspark.sql.window import Window

from databricks.kie.kie_evaluator import parse_json_markdown
from databricks.model_serving.types.pt_endpoint import BaseEndpoint, TileMTBatchEndpoint
from databricks.model_training.api.utils import check_if_table_exists, check_table_has_columns, get_spark

VALID_EXTENSIONS = {
    '.c', '.cpp', '.cs', '.css', '.go', '.html', '.java', '.js', '.json', '.md', '.php', '.py', '.rb', '.sh', '.tex',
    '.ts', '.txt', '.log', ''
}

MIN_LABELED_TRAIN_SAMPLES = 100
MIN_DOCS_TO_TRAIN = 50
MAX_DOCS_TO_TRAIN = 200  # for ai_query, we cannot support total_tokens yet, so set temp limit until supported

IGNORED_CACHE_ROOT = 'kie_cache_'

CHAT_TABLE_SCHEMA = StructType([
    StructField(
        "messages",
        ArrayType(StructType([StructField("role", StringType(), False),
                              StructField("content", StringType(), False)])), True)
])


def cache_on_serverless(func):

    @wraps(func)
    def wrapper(*args, **kwargs):
        result = func(*args, **kwargs)
        spark = get_spark()
        return spark.createDataFrame(result.toPandas())

    return wrapper


def cache_df(df: DataFrame, path: str) -> None:
    """Cache the dataframe at the specified path
    
    Args:
        df (DataFrame): Dataframe to be cached
        path (str): Table path at which to cache the dataframe
    """
    df.write.mode("overwrite").option("overwriteSchema", "true").saveAsTable(path)


def get_split_from_labeled(labeled_split_df: DataFrame, split: str, request: str, expected_response: str) -> DataFrame:
    # yapf: disable
    return (labeled_split_df
            .where(F.col('split') == split)
            .withColumn('request', F.col(request))
            .withColumn('expected_response', F.col(expected_response))
    )  # yapf: enable


@cache_on_serverless
def get_split_from_unlabeled(unlabeled_split_df: DataFrame, split: str) -> DataFrame:
    df = unlabeled_split_df.where(F.col('split') == split)
    if 'request' in df.columns:
        return df
    elif 'doc_uri' in df.columns:
        return read_documents_to_column(df, "doc_uri", "request")
    else:
        raise ValueError("unlabeled_split_df must have either 'doc_uri' or 'request' column")

def read_documents_to_column(df, path_column: str = "doc_uri", output_column: str = "request"):
    # Get unique file paths
    file_paths_df = df.select(path_column).distinct()

    # Create schema for the output DataFrame
    schema = StructType([
        StructField("file_path", StringType(), False),
        StructField("request", StringType(), True)
    ])

    # Function to process each pandas DataFrame batch
    def read_files(batch_iter):
        for pdf in batch_iter:
            result = []
            for path in pdf["doc_uri"]:
                try:
                    # Using pandas to read the file
                    with open(path, 'r', encoding='utf-8', errors='replace') as file:
                        content = file.read()
                except Exception:  # pylint: disable=broad-except
                    content = None
                result.append((path, content))
            yield pd.DataFrame(result, columns=["file_path", "request"])

    # Apply mapInPandas to read files
    path_content_df = file_paths_df.mapInPandas(read_files, schema)

    # Join back with original DataFrame
    return df.join(path_content_df, df[path_column] == path_content_df.file_path, "left").select(
    *df.columns + [path_content_df[output_column]])


def get_all_unlabeled(unlabeled_split_df: DataFrame) -> DataFrame:
    """Get all unlabeled data from the split df """
    if 'request' in unlabeled_split_df.columns:
        return unlabeled_split_df
    elif 'doc_uri' in unlabeled_split_df.columns:
        return read_documents_to_column(unlabeled_split_df, "doc_uri", "request")
    else:
        raise ValueError("unlabeled_split_df must have either 'doc_uri' or 'request' column")

def get_valid_files(directory: str) -> List[str]:
    """ Recursively get all files in the directory that match valid extensions
    """
    all_files = []
    for root, _, files in os.walk(directory):
        if os.path.basename(root).startswith(IGNORED_CACHE_ROOT):
            # This is a cache folder - skip it
            continue
        for f in files:
            if os.path.splitext(f)[1].lower() in VALID_EXTENSIONS:
                all_files.append(os.path.join(root, f))

    return all_files


def split_labeled_data(
    df: DataFrame,
    num_grounding_samples: int = 5,
    num_val_samples: int = 25,
    seed: int = 42,
) -> DataFrame:

    # Count the total sample size
    total_samples = df.count()

    # Allocate first to grounding, then val and training
    labeled_remaining = total_samples

    num_grounding_samples = min(num_grounding_samples, labeled_remaining)
    labeled_remaining -= num_grounding_samples

    # Default split to train
    df = df.withColumn('split', F.lit('train'))

    # Create random data splits
    # Order by rand so we can sample from it
    df = df.withColumn('rand', F.rand(seed=seed)).orderBy('rand')
    df = df.withColumn('row_id', F.row_number().over(Window.partitionBy("split").orderBy('rand')))

    # Split grounding and val data
    df = df.withColumn('split',
                       F.when(F.col('row_id') <= num_grounding_samples, F.lit('grounding')).otherwise(F.col('split')))
    df = df.withColumn('split',
                       F.when((F.col('row_id') > num_grounding_samples) & \
                              (F.col('row_id') <= num_grounding_samples + num_val_samples),
                              F.lit('val')).otherwise(F.col('split')))

    return df


def create_unlabeled_df_from_dataset(
    dataset: Optional[str],
) -> Tuple[DataFrame, List[str]]:
    all_files = get_valid_files(dataset)

    if not all_files:
        raise ValueError(f"No files found in {dataset} matching accepted file types: {','.join(VALID_EXTENSIONS)}")

    # Create a dataframe with the file paths
    spark = get_spark()
    schema = StructType([StructField('doc_uri', StringType(), True)])

    # Create DataFrame with schema
    df = spark.createDataFrame([(uri,) for uri in all_files], schema=schema).orderBy('doc_uri')
    return df, len(all_files)

def create_unlabeled_df_from_delta_table(
    delta_table: Optional[str],
    delta_table_text_column: Optional[str],
) -> Tuple[DataFrame, int]:
    spark = get_spark()
    df = (spark.read.table(delta_table)
          .select(delta_table_text_column)
          .withColumnRenamed(delta_table_text_column, 'request'))
    return df, df.count()

def split_unlabeled_data(
    dataset: Optional[str],
    delta_table: Optional[str],
    delta_table_text_column: Optional[str],
    num_val_samples: int = 25,
    num_grounding_samples: int = 5,
    seed: int = 42,
) -> DataFrame:
    if dataset:
        df, total_samples = create_unlabeled_df_from_dataset(dataset)
    elif delta_table:
        df, total_samples = create_unlabeled_df_from_delta_table(delta_table, delta_table_text_column)
    else:
        raise ValueError("Please provide one of dataset or delta_table")

    # Adjust sample counts to respect max percentages
    max_grounding = max(int(total_samples * 0.10), 1)  # 10% max for grounding
    max_val = max(int(total_samples * 0.20), 1)  # 20% max for validation

    # Calculate needed samples respecting maximums
    grounding_needed = min(num_grounding_samples, max_grounding)
    val_needed = min(num_val_samples, max_val)

    # Default split to unused
    df = df.withColumn('split', F.lit('unused'))

    # Order by rand so we can sample from it
    df = df.withColumn('rand', F.rand(seed=seed)).orderBy('rand')
    df = df.withColumn('row_id', F.row_number().over(Window.partitionBy("split").orderBy('rand')))

    # Split grounding and val data
    running_count = 0
    if grounding_needed > 0:
        df = df.withColumn('split',
                           F.when(F.col('row_id') <= grounding_needed, F.lit('grounding')).otherwise(F.col('split')))
        running_count += grounding_needed

    if val_needed > 0:
        df = df.withColumn(
            'split',
            F.when((F.col('row_id') > running_count) & (F.col('row_id') <= running_count + val_needed),
                   F.lit('val')).otherwise(F.col('split')))
        running_count += val_needed

    if dataset:
        return df.select('doc_uri', 'split').orderBy('doc_uri')
    else:
        return df.select('request', 'split')


def get_fewshot_examples(
    df: DataFrame,
    num_examples: int,
    response_format: Type[BaseModel],
    response_column: str = 'expected_response',
) -> List[Tuple[str, Dict[str, Any]]]:
    """
    Extract valid few-shot examples from the provided DataFrame.

    Args:
        df: DataFrame containing request/response pairs
        num_examples: Number of examples to extract
        response_format: Expected format for validation
        response_column: Column containing the expected response

    Returns:
        List of (request, response) tuples
    """
    examples = []
    for row in df.limit(num_examples).collect():
        try:
            response = json.loads(row[response_column])
            response_format(**response)  # Validate response
            examples.append((row['request'], response))
            if len(examples) >= num_examples:
                break
        except (json.JSONDecodeError, ValidationError):
            continue
    return examples

def filter_valid_json(df: DataFrame, column_name: str, schema: Type[BaseModel]) -> DataFrame:
    spark = get_spark()

    # Convert to pandas
    pdf = df.toPandas()

    # Validate each row
    valid_mask = pdf[column_name].apply(lambda x: _is_valid_json(x, schema))

    # Filter and convert back to Spark
    filtered_pdf = pdf[valid_mask]
    return spark.createDataFrame(filtered_pdf, df.schema)

def _is_valid_json(maybe_json_str: str, schema: Type[BaseModel]) -> bool:
    try:
        schema(**json.loads(maybe_json_str))
        return True
    except:  # pylint: disable=bare-except
        return False

def _make_training_row(
    prompt: str,
    request: str,
    response: str,
    response_format: Type[BaseModel],
) -> dict:
    # We reload and then dump the json here so that the order of the keys is preserved based on the provided schema
    # This matters because ai_query may return the keys in a different order. Once ai_query has fixed this bug, this
    # code can be removed.
    try:
        json_loaded = json.loads(parse_json_markdown(response))
    except json.JSONDecodeError as e:
        raise ValueError(
            f"Failed to parse JSON from response: {response}. " +
            "Please make sure to filter invalid responses before this function.") from e

    try:
        schema_loaded = response_format(**json_loaded)
    except ValidationError as e:
        raise ValueError(
            f"Failed to validate schema from response: {response}. " +
            "Please make sure to filter invalid responses before this function.") from e

    json_dumped = schema_loaded.model_dump_json()

    user_message = prompt + '\n\nDocument:\n\n' + request.replace('"', '').replace('“', '').replace(
        '”', '') + '\n\n' + prompt
    messages = [{
        'role': 'user',
        'content': user_message
    }, {
        'role': 'assistant',
        'content': json_dumped,
    }]
    messages_line = {'messages': messages}
    return messages_line


def process_for_sampling(df: DataFrame, seed: int = 42) -> DataFrame:
    spark = get_spark()
    df_shuffled = df.orderBy(rand(seed=seed))
    df_with_id = df_shuffled.withColumn('id', monotonically_increasing_id())
    df_with_id_static = spark.createDataFrame(df_with_id.toPandas())
    return df_with_id_static


def batched_iterator(df: DataFrame, batch_size: int = 1000) -> Iterable[pd.DataFrame]:
    current_max_index = 0
    while current_max_index < df.count():
        current_df = df.where((df.id >= current_max_index) & (df.id < current_max_index + batch_size))
        if 'request' in current_df.columns:
            yield current_df.toPandas()
        elif 'doc_uri' in current_df.columns:
            yield read_documents_to_column(current_df, "doc_uri", "request").toPandas()
        else:
            raise ValueError("df must have either 'doc_uri' or 'request' column")
        current_max_index += batch_size


def write_batch_to_cache(prompt: str,
                         documents: List[str],
                         responses: List[str],
                         response_format: Type[BaseModel],
                         file_path: str):
    lines = []
    for request, response in zip(documents, responses):
        if not _is_valid_json(response, response_format):
            continue

        training_row = _make_training_row(
            prompt=prompt,
            request=request,
            response=response,
            response_format=response_format
        )
        lines.append(json.dumps(training_row))

    with open(file_path, 'a', encoding='utf-8') as fh:
        fh.write('\n'.join(lines) + '\n' if lines else '')

def format_dataframe_for_training(
    df: DataFrame,
    prompt_column: str,
    request_column: str,
    response_column: str,
    response_format: Type[BaseModel],
) -> DataFrame:
    spark = get_spark()

    pdf = df.toPandas()

    formatted_rows = []
    for _, row in pdf.iterrows():
        prompt = row[prompt_column]
        request = row[request_column]
        response = row[response_column]

        training_row = _make_training_row(
            prompt=prompt,
            request=request,
            response=response,
            response_format=response_format,
        )

        formatted_rows.append(training_row)

    return spark.createDataFrame(formatted_rows, schema=CHAT_TABLE_SCHEMA)

def filter_format_and_write_training_dataframe(
    df: DataFrame,
    prompt_column: str,
    request_column: str,
    response_column: str,
    response_format: Type[BaseModel],
    file_path: str,
    is_jsonl: bool,
) -> None:
    df = filter_valid_json(df, response_column, response_format)
    formatted_df = format_dataframe_for_training(
        df,
        prompt_column,
        request_column,
        response_column,
        response_format=response_format,
    )

    if is_jsonl:
        formatted_df.toPandas().to_json(
            file_path,
            orient='records',
            lines=True
        )
    else:
        formatted_df.write.format("delta").mode("overwrite").saveAsTable(file_path)


def create_training_data(
    unlabeled_split_df: DataFrame,
    prompt: str,
    response_format: Type[BaseModel],
    train_jsonl_cache_path: str,
    train_table_path: Optional[str],
    endpoint: BaseEndpoint,
    token_budget: int,
    min_docs_to_train: int = MIN_DOCS_TO_TRAIN,
    use_ai_query: bool = False,
) -> Optional[Tuple[str, int]]:

    train_df = unlabeled_split_df.where((F.col('split') == 'train') | (F.col('split') == 'unused'))

    if train_df.count() < min_docs_to_train:
        return None

    df_path = train_table_path or train_jsonl_cache_path

    if use_ai_query:
        # Read training documents into the table and save it out
        train_df = get_all_unlabeled(train_df.limit(MAX_DOCS_TO_TRAIN))
        train_df = get_spark().createDataFrame(train_df.toPandas())

        # If using a TileMTBatchEndpoint, we need to validate that system_prompt exists
        if isinstance(endpoint, TileMTBatchEndpoint) and not endpoint.system_config.system_prompt_header:
            raise ValueError("system_prompt_header must be set in the system config for TileMTBatchEndpoint")

        # Write to temporary view
        train_df.createOrReplaceTempView("train_df")

        output_column = 'response'
        result_df = endpoint.generate_ai_query(
            data_path="train_df",
            system_prompt=prompt,
            response_format=response_format,
            output_column=output_column,
        )
        print(f"Finished generating {result_df.count()} responses from ai_query")

        filter_format_and_write_training_dataframe(
            result_df,
            'system_prompt',
            'request',
            output_column,
            response_format,
            df_path,
            train_table_path is None
        )

        print("Wrote training data to path: ", df_path)
        # Note that we don't have support for token budget yet, so we are just returning 0
        return df_path, 0

    ready_for_sampling = process_for_sampling(train_df)
    token_budget_remaining = token_budget
    data_records = []

    for batch in batched_iterator(ready_for_sampling):
        documents = batch['request'].tolist()

        responses = endpoint.generate_batch(
            prompts=documents,
            system_prompt=prompt,
            total_max_tokens=token_budget_remaining,
            response_format=response_format,
            show_progress=False,
        )

        actual_responses = [r.response for r in responses]

        for doc, resp in zip(documents, actual_responses):
            record = {
                "prompt": prompt,
                "document": doc,
                "response": resp,
            }
            data_records.append(record)

        token_budget_remaining -= sum(r.total_tokens for r in responses)

    spark = get_spark()
    result_df = spark.createDataFrame(data_records)

    filter_format_and_write_training_dataframe(
        result_df,
        'prompt',
        'document',
        'response',
        response_format,
        df_path,
        train_table_path is None,
    )

    print("Wrote training data to path: ", df_path)
    return df_path, token_budget - token_budget_remaining


def validate_row_count(
    df: DataFrame,
    required_count: int = 10,
    recommended_count: int = 1000,
) -> None:
    row_count = df.count()
    if row_count < required_count:
        raise ValueError(f'Insufficient data. Found {row_count} rows, expected at least {required_count}.')
    if row_count < recommended_count:
        print(f'Warning: Found {row_count} rows, recommended at least {recommended_count} for best results.')


def to_row_based_format(column_based_data: Dict[str, List[str]]) -> List[Dict[str, str]]:
    """
    Converts column-based data to row-based format.

    Args:
        column_based_data (dict): A dictionary where keys are column names and values are lists of column values.

    Returns:
        list: A list of dictionaries where each dictionary represents a row of data.
    """
    row_based_data = []
    num_rows = len(next(iter(column_based_data.values())))

    if len(set(len(values) for values in column_based_data.items())) != 1:
        raise ValueError('column-based data must have the same number of rows for all columns')

    for i in range(num_rows):
        row = {key: value[i] for key, value in column_based_data.items()}
        row_based_data.append(row)
    return row_based_data


def to_column_based_format(row_based_data: List[Dict[str, str]]) -> Dict[str, List[str]]:
    """
    Converts row-based data to column-based format.

    Args:
        row_based_data (list): A list of dictionaries where each dictionary represents a row of data.

    Returns:
        dict: A dictionary where keys are column names and values are lists of column values.
    """
    if not row_based_data:
        return {}

    column_based_data = {key: [] for key in row_based_data[0].keys()}
    for row in row_based_data:
        if set(row.keys()) != set(column_based_data.keys()):
            raise ValueError('all rows must have the same keys')

        for key, value in row.items():
            column_based_data[key].append(value)
    return column_based_data


def read_from_table_with_columns(
    table_path: str,
    columns: Tuple[str],
    count: int,
    filter_null_value: bool = False,
) -> List[Dict[str, Any]]:
    """
    Reads specified columns from a table and returns a DataFrame with a limited number of rows.

    Args:
        table_path (str): The path to the table.
        columns (List[str]): The list of columns to read.
        count (int): The number of rows to read.
        filter_null_value (bool): Whether to filter out rows with null values in the specified columns.

    Returns:
        list: A list of dictionaries where each dictionary represents a row of the DataFrame.
    """
    if not check_if_table_exists(table_path):
        raise ValueError(f'{table_path} is not a valid table path. Please check that your table exists.')
    if not check_table_has_columns(table_path, columns):
        raise ValueError(f'{table_path} does not have expected columns: {columns}')

    spark = get_spark()
    df = spark.read.table(table_path).select(*columns)
    if filter_null_value:
        df = df.filter(" AND ".join(f"{col} IS NOT NULL" for col in columns))
    df = df.limit(count).collect()

    return [datum.asDict() for datum in df]

def read_from_jsonl(file_path: str, count: Optional[int]= None) -> List[Dict[str, Any]]:
    """
    Reads a JSONL file and returns a specified number of rows.

    Args:
        file_path (str): The path to the JSONL file.
        count (int): Optional number of rows to read.

    Returns:
        list: A list of dictionaries where each dictionary represents a row of the JSONL file.
    """
    json_obj = pd.read_json(path_or_buf=file_path, lines=True, nrows=count)
    return json_obj.to_dict(orient='records') # pylint: disable=no-member

def partition_data(
    data: List[Any],
    min_test_data_size: int = 10,
    test_set_ratio: float = 0.5,
) -> Tuple[List[Any], List[Any]]:
    """Randomly split data to test data and train data."""
    data_to_partition = deepcopy(data)
    example_data_size = len(data_to_partition)
    if example_data_size <= min_test_data_size:
        return [], data_to_partition

    random.shuffle(data_to_partition)
    split_index = int(example_data_size * test_set_ratio)
    return data_to_partition[split_index:], data_to_partition[:split_index]



"""Model that selects the best response from a set of responses based on a value model."""
from concurrent.futures import ThreadPoolExecutor
from typing import Callable, Optional

from databricks.kie.research_common.reward_models import ContextRewardModel, RewardModel, make_reward_model
from databricks.kie.t2t_models.base_t2t_model import BaseT2TModel
from databricks.kie.t2t_models.t2t_model_registry import model_registry
from databricks.kie.t2t_schema import T2TSystemParams

DEFAULT_BEST_OF_N = 8
DEFAULT_INNER_MODEL_NAME = "t2t_baseline_model"


class BestOfNModel(BaseT2TModel):
    """Model that selects the best response from a set of responses based on a value model."""

    def __init__(self, inner_model: BaseT2TModel, reward_model: RewardModel, best_of_n: int = DEFAULT_BEST_OF_N):
        self.inner_model = inner_model
        self.reward_model = reward_model
        self.best_of_n = best_of_n

    @staticmethod
    def create_from_system_param(system_param: T2TSystemParams, **kwargs) -> Optional[BaseT2TModel]:
        best_of_n = kwargs.pop("best_of_n", DEFAULT_BEST_OF_N)
        value_model_name = kwargs.pop("value_model_name", "context")
        value_model_params = kwargs.pop("value_model_params", {})
        inner_model_name = kwargs.pop("inner_model_name", DEFAULT_INNER_MODEL_NAME)

        reward_model = make_reward_model(value_model_name, **value_model_params)
        inner_model_constructor: Callable[[T2TSystemParams], BaseT2TModel] = model_registry.get(inner_model_name)
        inner_model = inner_model_constructor(system_param, **kwargs)
        return BestOfNModel(inner_model=inner_model, reward_model=reward_model, best_of_n=best_of_n)

    def __call__(self, model_input: str) -> str:
        return select_best_response(self.inner_model, model_input, self.reward_model, self.best_of_n)


def select_best_response(model: BaseT2TModel,
                         prompt: str,
                         reward_model: Optional[RewardModel] = None,
                         n: int = DEFAULT_BEST_OF_N) -> str:
    if reward_model is None:
        reward_model = ContextRewardModel()
    results = []
    values = []

    # parallelize this
    def get_result_and_value(_):
        result = model(prompt)
        value = reward_model(prompt, result)
        return result, value

    with ThreadPoolExecutor(max_workers=DEFAULT_BEST_OF_N) as executor:
        results_and_values = executor.map(get_result_and_value, range(n))

    results, values = zip(*results_and_values)

    return max(zip(results, values), key=lambda x: x[1])[0]



"""Module containing the utility functions related to reward model."""
from typing import Optional

from pydantic import BaseModel, Field

from databricks.kie.inference_utils import DEFAULT_NUM_RETRIES
from databricks.kie.t2t_utils import (create_chat_completions_messages_from_instruction,
                                      create_structured_outputs_using_cot)

PROMPT_COMPLEXITY_RATING_DESCRIPTION = """Rate the complexity of the prompt:
(1-very easy), (2-easy), (3-normal), (4-complex), (5-very complex).
For example, 
    - (1-very easy) prompt might be a simple yes or no question, 
    - (3-normal) prompt might require a bit of thought or some specific memorable knowledge 
    - (5-very complex) prompt might require complex reasoning and/or breaking down the question 
        into multiple subquestions and/or require expertise of PhD level human.
"""

PROMPT_COMPLEXITY_RATING_SYSTEM_PROMPT = """
You are an advanced annotator AI able to simulate the expert judgement of a top level human annotator. 
Your task is to analyze a prompt given by the user and label its complexity.

## Input
You will receive a user prompt in the under the following header format: ## prompt_to_analyze

## Task
You goal is to label the complexity & difficulty of the prompt.

## Output
You should output the complexity rating of the prompt as rating integer between 1 and 5.
1 being the easiest and 5 being the hardest.
"""

REWARD_RATING_DESCRIPTION = """Rate the quality of the response:
(1-very bad), (2-bad), (3-normal), (4-good), (5-very good).
For example, 
    - (1-very bad) response might be a response that is not helpful, incorrect, or not relevant to the prompt.
    - (3-normal) response might be a response that is helpful, correct, and relevant to the prompt.
    - (5-very good) response might be a response that is helpful, correct, and relevant to the prompt.
"""

REWARD_RATING_SYSTEM_PROMPT = """
You are an advanced annotator AI able to simulate the expert judgement of a top level human annotator. 
Your task is to analyze a response given by the user and label its quality.

## Input
You will receive a user prompt and a response.
Prompt will be provided under the following header format: ## prompt_to_analyze
Response will be provided under the following header format: ## response_to_analyze

## Task
You goal is to label the quality of the response.

## Output
You should output the quality rating of the response as rating integer between 1 and 5.
1 being the worst and 5 being the best.
"""

COMPLEXITY_RATING_USER_PROMPT_TEMPLATE = """
## prompt_to_analyze: '{prompt}'
"""

REWARD_RATING_USER_PROMPT_TEMPLATE = """
## prompt_to_analyze: '{prompt}'
## response_to_analyze: '{response}'
"""

DEFAULT_CONTEXT_REWARD_MODEL_ID = "gpt-4o-2024-08-06-text2text"


class PromptComplexityRating(BaseModel):
    rating: int = Field(description=PROMPT_COMPLEXITY_RATING_DESCRIPTION)


class RewardRating(BaseModel):
    rating: int = Field(description=REWARD_RATING_DESCRIPTION)


def generate_prompt_complexity_rating(prompt: str, model_id: Optional[str] = DEFAULT_CONTEXT_REWARD_MODEL_ID) -> int:
    """
    Generate a rating based reward value for a given prompt and response.

    Args:
        prompt (str): The input prompt to evaluate along with the response.
        model_id (str): The model id to use for generating the reward value.

    Returns:
        int: The reward value between 1 and 10.
    """
    messages = create_chat_completions_messages_from_instruction(
        instruction=PROMPT_COMPLEXITY_RATING_SYSTEM_PROMPT,
        inp=COMPLEXITY_RATING_USER_PROMPT_TEMPLATE.format(prompt=prompt))
    return create_structured_outputs_using_cot(PromptComplexityRating,
                                               messages,
                                               model_id,
                                               num_retries=DEFAULT_NUM_RETRIES).rating


def generate_reward_rating(prompt: str,
                           response: str,
                           model_id: Optional[str] = DEFAULT_CONTEXT_REWARD_MODEL_ID) -> int:
    """
    Generate a rating based reward value for a given prompt and response.

    Args:
        prompt (str): The input prompt to evaluate along with the response.
        response (str): The response to generate the reward value for.
        model_id (str): The model id to use for generating the reward value.

    Returns:
        int: The reward value between 1 and 10.
    """
    messages = create_chat_completions_messages_from_instruction(instruction=REWARD_RATING_SYSTEM_PROMPT,
                                                                 inp=REWARD_RATING_USER_PROMPT_TEMPLATE.format(
                                                                     prompt=prompt, response=response))
    return create_structured_outputs_using_cot(RewardRating, messages, model_id, num_retries=DEFAULT_NUM_RETRIES).rating



"""Utils for working with model serving"""

from enum import Enum
from typing import Dict, NamedTuple, Optional

from mlflow import MlflowClient

MODEL_CONTEXT_LENGTH = 131072


class ModelSpecs(NamedTuple):
    name: str
    endpoint: Optional[str]
    cost_per_hour: float  # This is measured in DBUs instead of dollars, since customers may have different DBU rates
    ft_model_name: str
    is_hosted: bool = True
    uc_schema: Optional[str] = None
    version: Optional[str] = None
    cost_per_m_ft_tokens: Optional[float] = None
    alias: Optional[str] = None
    throughput_tok_per_sec: Optional[int] = None


class ModelSpecKey(Enum):
    """Enumeration of model specification keys for model serving."""
    # 70B variants
    BALANCED = 'balanced'
    LLAMA_3_3_70B_INSTRUCT = 'llama-3-3-70b-instruct'

    # 8B variants
    COST_OPTIMIZED_MEDIUM = 'cost-optimized-medium'
    LLAMA_3_1_8B_INSTRUCT = 'llama-3-1-8b-instruct'

    # 3B variants
    COST_OPTIMIZED_SMALL = 'cost-optimized-small'
    LLAMA_3_2_3B_INSTRUCT = 'llama-3-2-3b-instruct'


def get_latest_model_version(model_name: str) -> str:
    """Get the latest version of a model using Mlflow

    Args:
        model_name (str): The name of the model

    Returns:
        str: The latest version of the model
    """
    client = MlflowClient()
    model_versions = client.search_model_versions(f"name='{model_name}'")

    if not model_versions:
        raise ValueError(f"No model versions found for model '{model_name}'")

    version_numbers = [int(model.version) for model in model_versions]
    most_recent_version = max(version_numbers)
    return str(most_recent_version)


def create_model_specs(name: str,
                       endpoint: Optional[str],
                       cost_per_hour: float,
                       ft_model_name: str,
                       uc_schema: Optional[str] = None,
                       is_hosted: bool = True,
                       cost_per_m_ft_tokens: Optional[float] = None,
                       throughput_tok_per_sec: Optional[int] = None,
                       alias: Optional[str] = None) -> ModelSpecs:
    """Creates a model specification with optional automatic version lookup.

    Args:
        name (str): Model spec name.
        endpoint (Optional[str]): PayGo serving endpoint.
        cost_per_hour (float): Cost per hour in DBUs.
        ft_model_name (str): Finetuning API model name.
        uc_schema (Optional[str]): Unity Catalog schema name.
        is_hosted (bool, optional): Whether the model is hosted. Defaults to True.
        cost_per_m_ft_tokens (Optional[float], optional): Cost per million fine-tuned tokens.
        throughput_tok_per_sec (Optional[int], optional): Model throughput in tokens per second.
        alias (Optional[str], optional): Friendly alias for the model.

    Returns:
        ModelSpecs: A structured model specification.
    """
    version = None
    if uc_schema:
        try:
            version = get_latest_model_version(uc_schema)
        except ValueError as e:
            print(f"Failed to get latest model version for {uc_schema}: {e}")

    return ModelSpecs(name=name,
                      endpoint=endpoint,
                      cost_per_hour=cost_per_hour,
                      ft_model_name=ft_model_name,
                      is_hosted=is_hosted,
                      uc_schema=uc_schema,
                      version=version,
                      cost_per_m_ft_tokens=cost_per_m_ft_tokens,
                      throughput_tok_per_sec=throughput_tok_per_sec,
                      alias=alias)


BASE_MODELS: Dict[str, ModelSpecs] = {
    # 70B variants
    ModelSpecKey.LLAMA_3_3_70B_INSTRUCT.value:
        create_model_specs(name="llama-3-3-70b-instruct",
                           endpoint="databricks-meta-llama-3-3-70b-instruct",
                           cost_per_hour=342.857,
                           ft_model_name="meta-llama/Llama-3.3-70B-Instruct",
                           uc_schema="system.ai.llama_v3_3_70b_instruct",
                           throughput_tok_per_sec=670,
                           alias="elephant"),

    # 8B variants
    ModelSpecKey.LLAMA_3_1_8B_INSTRUCT.value:
        create_model_specs(name="llama-3-1-8b-instruct",
                           endpoint="databricks-meta-llama-3-1-8b-instruct",
                           cost_per_hour=106.0,
                           ft_model_name="meta-llama/Meta-Llama-3.1-8B-Instruct",
                           uc_schema="system.ai.llama_v3_1_8b_instruct",
                           cost_per_m_ft_tokens=4.0,
                           throughput_tok_per_sec=19000,
                           alias="zebra"),

    # 3B variants
    ModelSpecKey.LLAMA_3_2_3B_INSTRUCT.value:
        create_model_specs(name="llama-3-2-3b-instruct",
                           endpoint=None,
                           cost_per_hour=92.857,
                           ft_model_name="meta-llama/Llama-3.2-3B-Instruct",
                           uc_schema="system.ai.llama_v3_2_3b_instruct",
                           is_hosted=False,
                           cost_per_m_ft_tokens=2.50,
                           throughput_tok_per_sec=22000,
                           alias="squirrel"),
}


def get_model_specs():
    """Retrieves all available model specifications.

    Returns:
        Dict[ModelSpecKey, ModelSpecs]: A dictionary mapping model keys to their corresponding model specifications.
    """
    return {
        **BASE_MODELS,
        ModelSpecKey.BALANCED.value: BASE_MODELS[ModelSpecKey.LLAMA_3_3_70B_INSTRUCT.value],
        ModelSpecKey.COST_OPTIMIZED_MEDIUM.value: BASE_MODELS[ModelSpecKey.LLAMA_3_1_8B_INSTRUCT.value],
        ModelSpecKey.COST_OPTIMIZED_SMALL.value: BASE_MODELS[ModelSpecKey.LLAMA_3_2_3B_INSTRUCT.value],
    }



"""Functions for creating evaluation criteria."""

import json
from typing import List, Optional, Tuple

from pydantic import BaseModel, Field

from databricks.kie.inference_utils import generate_base_model_using_chat_completion_messages
from databricks.kie.t2t_schema import PreferenceData

_EVAL_CRITERIA_GENERATOR_SYSTEM_PROMPT = (
    "You are an expert at user instruction analysis. "
    "Given an instruction for any task provided by the user, return a JSON list of criteria."
    "The criteria will be used to score an LLM on how well it followed the instructions given a specific input."
    "Provide detailed granular evaluation criteria on fomatting, style, and correctness."
    "Instruction to analyze is provided in in user message following 'Instruction' header."
    "Provide detailed granular evaluation criteria on fomatting, style, and correctness.")

_EVAL_CRITERIA_GENERATOR_SYSTEM_PROMPT_WITH_EXAMPLES = (
    "You are an expert at user instruction analysis. "
    "Given an instruction for any task provided by the user and a JSON containing ground truth request"
    " and response pairs, return a JSON list of criteria."
    "The criteria will be used to score an LLM on how well it followed the instructions given a specific input."
    "Instruction to analyze is provided in in user message following 'Instruction' header, "
    "and examples are provided in user message following 'Ground Truth Examples' header."
    "Provide detailed granular evaluation criteria on fomatting, style, and correctness.")

_EVAL_CRITERIA_GENERATOR_SYSTEM_PROMPT_FROM_PREFERENCE = (
    "You are an expert at error analysis. "
    "Given an instruction for a task provided by the user and a JSON containing the input, incorrect "
    "output, and correct output, return a JSON list of differences between the incorrect and"
    "correct output. "
    "The criteria will be used to score an LLM on how well it followed the instructions given a specific input. "
    "Instruction to analyze is provided in in user message following 'Instruction' header, "
    "and examples are provided in user message following 'Examples' header. "
    "Provide specific evaluation criteria about the differences between the incorrect and correct output. "
    "Only provide the differences, not the similarities. If the criteria is the same, do not repeat it. "
    "Make sure the differences are between the incorrect and correct output, not just a difference between the "
    "input and incorrect output. "
    "Generalize the differences as criteria to apply to all examples")

_DEDUPE_EVAL_CRITERIA_GENERATOR_SYSTEM_PROMPT = (
    "You are an expert at deduplicating lists of evaluation criteria. "
    "Given an instruction for a task provided by the user and a list of the old criteria and a list of the "
    "new criteria, return a JSON list of criteria combining the two. "
    "The criteria will be used to score an LLM on how well it followed the instructions given a specific input. "
    "Instruction to analyze is provided in in user message following 'Instruction' header, "
    "the old criteria are provided in user message following 'Old criteria' header, "
    "and the new criteria are provided in user message following 'New criteria' header. "
    "Do not repeat criteria that are the same. If there are contradictory criteria between the old criteria "
    "and the new criteria, the new criteria takes priority.")


class EvaluationCriteria(BaseModel):
    criteria: List[str] = Field(description="List of evaluation criteria that the response must satisfy.")


def generate_evaluation_criteria(instruction: str, examples: Optional[List[Tuple[str, str]]] = None) -> List[str]:
    """Generates list of evaluation criteria based on the given instruction and examples.

    Args:
        instruction (str): The instruction to generate evaluation criteria for.
        examples (list, optional): A list of tuple (input, output) containing ground truth request and response pairs.
            Defaults to None.

    Returns:
        list: A list of generated evaluation criteria.
    """
    if examples:
        examples = [{"request": example[0], "response": example[1]} for example in examples]

    messages = [{
        "role":
            "system",
        "content":
            _EVAL_CRITERIA_GENERATOR_SYSTEM_PROMPT_WITH_EXAMPLES if examples else _EVAL_CRITERIA_GENERATOR_SYSTEM_PROMPT
    }, {
        "role":
            "user",
        "content": (f"Instruction: {instruction}\n\n Ground Truth Examples: {json.dumps(examples)}"
                    if examples else f"Instruction: {instruction}")
    }]

    return generate_base_model_using_chat_completion_messages(messages,
                                                              EvaluationCriteria,
                                                              model_id="gpt-4o-2024-08-06-text2text").criteria


def generate_evaluation_criteria_from_preference(instruction: str, preference_data: List[PreferenceData]) -> List[str]:
    """Generates list of evaluation criteria based on the given instruction and preference data.

    Args:
        instruction (str): The instruction to generate evaluation criteria for.
        preference_data (list): A list of preference data.

    Returns:
        list: A list of generated evaluation criteria.
    """
    examples = [{
        "input": example.input,
        "incorrect output": example.rejected_response,
        "correct output": example.preferred_response,
    } for example in preference_data]

    messages = [{
        "role": "system",
        "content": _EVAL_CRITERIA_GENERATOR_SYSTEM_PROMPT_FROM_PREFERENCE
    }, {
        "role": "user",
        "content": f"Instruction: {instruction}\n\nExamples: {json.dumps(examples)}"
    }]

    return generate_base_model_using_chat_completion_messages(messages,
                                                              EvaluationCriteria,
                                                              model_id="gpt-4o-2024-08-06-text2text").criteria


def dedupe_evaluation_criteria(instruction: str, old_criteria: List[str], new_criteria: List[str]) -> List[str]:
    """Generates list of evaluation criteria based on the given instruction and criteria.

    The generated list of evaluation criteria combines the old criteria and new criteria, deduplicating similar
    criteria and giving preference to the new criteria if conflicts exist.

    Args:
        instruction (str): The instruction to generate evaluation criteria for.
        old_criteria (list): A list of old evaluation criteria.
        new_criteria (list): A list of new evaluation criteria.

    Returns:
        list: A list of deduped evaluation criteria.
    """

    messages = [{
        "role": "system",
        "content": _DEDUPE_EVAL_CRITERIA_GENERATOR_SYSTEM_PROMPT
    }, {
        "role":
            "user",
        "content": (f"Instruction: {instruction}\n\nOld criteria: {json.dumps(old_criteria)}\n\n"
                    f"New criteria: {json.dumps(new_criteria)}")
    }]

    return generate_base_model_using_chat_completion_messages(messages,
                                                              EvaluationCriteria,
                                                              model_id="gpt-4o-2024-08-06-text2text").criteria


def add_evaluation_criteria_from_preference_data(
    instruction: str,
    evaluation_criteria: List[str],
    preference_data: List[PreferenceData],
) -> List[str]:
    """Generates list of evaluation criteria based on the given instruction, old criteria, and new examples.

    The examples are first used to generate a set of new criteria. The old critera and new criteria are then combined.

    Args:
        instruction (str): The instruction to generate evaluation criteria for.
        evaluation_criteria (list): A list of evaluation criteria.
        preference_data (list): A list of preference data.

    Returns:
        list: A list of updated evaluation criteria.
    """
    new_criteria = generate_evaluation_criteria_from_preference(instruction=instruction,
                                                                preference_data=preference_data)
    final_criteria = dedupe_evaluation_criteria(
        instruction=instruction,
        old_criteria=evaluation_criteria,
        new_criteria=new_criteria,
    )
    return final_criteria
